#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import numpy as np
from collections import OrderedDict
from .strategy import strategy_registry, TuneStrategy
from ..utils import logger

@strategy_registry
class BasicTuneStrategy(TuneStrategy):
    """The basic tuning strategy which tunes the low precision model with below order.

    1. modelwise tuning for all quantizable ops.
    2. fallback tuning from bottom to top to decide the priority of which op has biggest impact
       on accuracy.
    3. incremental fallback tuning by fallbacking multiple ops with the order got from #2.

    Args:
        model (object):                        The FP32 model specified for low precision tuning.
        conf (Class):                          The Conf class instance initialized from user yaml
                                               config file.
        q_dataloader (generator):              Data loader for calibration, mandatory for
                                               post-training quantization.
                                               It is iterable and should yield a tuple (input,
                                               label) for calibration dataset containing label,
                                               or yield (input, _) for label-free calibration
                                               dataset. The input could be a object, list, tuple or
                                               dict, depending on user implementation, as well as
                                               it can be taken as model input.
        q_func (function, optional):           Reserved for future use.
        eval_dataloader (generator, optional): Data loader for evaluation. It is iterable
                                               and should yield a tuple of (input, label).
                                               The input could be a object, list, tuple or dict,
                                               depending on user implementation, as well as it can
                                               be taken as model input. The label should be able
                                               to take as input of supported metrics. If this
                                               parameter is not None, user needs to specify
                                               pre-defined evaluation metrics through configuration
                                               file and should set "eval_func" parameter as None.
                                               Tuner will combine model, eval_dataloader and
                                               pre-defined metrics to run evaluation process.
        eval_func (function, optional):        The evaluation function provided by user.
                                               This function takes model as parameter, and
                                               evaluation dataset and metrics should be
                                               encapsulated in this function implementation and
                                               outputs a higher-is-better accuracy scalar value.

                                               The pseudo code should be something like:

                                               def eval_func(model):
                                                    input, label = dataloader()
                                                    output = model(input)
                                                    accuracy = metric(output, label)
                                                    return accuracy
        dicts (dict, optional):                The dict containing resume information.
                                               Defaults to None.

    """

    def __init__(self, model, conf, q_dataloader, q_func=None,
                 eval_dataloader=None, eval_func=None, dicts=None, q_hooks=None):
        super(
            BasicTuneStrategy,
            self).__init__(
            model,
            conf,
            q_dataloader,
            q_func,
            eval_dataloader,
            eval_func,
            dicts,
            q_hooks)

    def next_tune_cfg(self):
        """The generator of yielding next tuning config to traverse by concrete strategies
           according to last tuning result.

        """
        # Model wise tuning
        op_cfgs = {}
        best_cfg = None
        if len(self.metric_name) == 1 or self.metric_weight is not None:
            best_acc = float('-inf') if self.higher_is_better else float('inf')
        else:
            best_acc = [float('-inf') if higher_is_better else float('inf') for \
                higher_is_better in self.metric_criterion]

        logger.debug("Start basic strategy by model-wise tuning.")
        for i, iterations in enumerate(self.calib_iter):
            op_cfgs['calib_iteration'] = int(iterations)
            op_cfgs['calib_sampling_size'] = int(self.calib_sampling_size[i])

            for combined_cfg in self.combined_model_wise_quant_cfgs:
                op_cfgs['op'] = OrderedDict()
                for op, op_cfg in self.opwise_quant_cfgs.items():
                    if op[1] in combined_cfg.keys() and len(op_cfg) > 0:
                        op_cfgs['op'][op] = copy.deepcopy(
                            self._get_common_cfg(combined_cfg[op[1]], op_cfg))
                    else:
                        op_cfgs['op'][op] = copy.deepcopy(
                            self.opwise_tune_cfgs[op][0])

                yield op_cfgs
                acc, _ = self.last_tune_result
                if not isinstance(acc, list) and ((self.higher_is_better and acc >= best_acc) \
                    or (not self.higher_is_better and acc <= best_acc)):
                    best_acc = acc
                    best_cfg = copy.deepcopy(op_cfgs)
                elif len(self.metric_name) > 1 and self.metric_weight is not None:
                    acc = np.mean(np.array(acc) * self.metric_weight)
                    if (self.higher_is_better and acc >= best_acc) or \
                        (not self.higher_is_better and acc <= best_acc):
                        best_acc = acc
                        best_cfg = copy.deepcopy(op_cfgs)
                elif len(self.metric_name) > 1 and self.metric_weight is None:
                    if all([acc_i >= best_i if higher_is_better else acc_i <= best_i for \
                        acc_i, best_i, higher_is_better in \
                        zip(acc, best_acc, self.metric_criterion)]):
                        best_acc = acc
                        best_cfg = copy.deepcopy(op_cfgs)

        if best_cfg is not None:
            fallback_dtypes = []
            for data_type in ["bf16", "fp16", "fp32"]:
                for op_type, tune_space in self.modelwise_tune_space.items():
                    if (data_type in tune_space["activation"]["dtype"] and
                        data_type not in fallback_dtypes):
                        fallback_dtypes.append(data_type)

            for fallback_dtype in fallback_dtypes:
                logger.debug(
                    "Continue basic strategy by sorting opwise {} fallback priority.".format
                    (fallback_dtype))
                ops_acc = OrderedDict()
                for op, configs in reversed(self.opwise_tune_cfgs.items()):
                    op_cfgs = copy.deepcopy(best_cfg)
                    for cfg in configs:
                        if ('weight' in cfg and fallback_dtype == cfg['weight']['dtype']) \
                           or ('weight' not in cfg
                               and fallback_dtype == cfg['activation']['dtype']):
                            if op_cfgs['op'][op]['weight'] is not None:
                                op_cfgs['op'][op]['weight'].clear()
                                op_cfgs['op'][op]['weight']['dtype'] = fallback_dtype
                            assert cfg['activation']['dtype'] == fallback_dtype
                            op_cfgs['op'][op]['activation'].clear()
                            op_cfgs['op'][op]['activation']['dtype'] = fallback_dtype
                    yield op_cfgs
                    acc, _ = self.last_tune_result
                    ops_acc[op] = acc

                op_cfgs = copy.deepcopy(best_cfg)
                if ops_acc is not None:
                    ordered_ops = sorted(ops_acc.keys(), key=lambda key: ops_acc[key],
                                         reverse=self.higher_is_better)
                    for op in ordered_ops:
                        old_cfg = copy.deepcopy(op_cfgs['op'][op])
                        for cfg in self.opwise_tune_cfgs[op]:
                            if ('weight' in cfg and fallback_dtype == cfg['weight']['dtype']) \
                               or ('weight' not in cfg
                                   and fallback_dtype == cfg['activation']['dtype']):
                                if op_cfgs['op'][op]['weight'] is not None:
                                    op_cfgs['op'][op]['weight'].clear()
                                    op_cfgs['op'][op]['weight']['dtype'] = fallback_dtype
                                assert cfg['activation']['dtype'] == fallback_dtype
                                op_cfgs['op'][op]['activation'].clear()
                                op_cfgs['op'][op]['activation']['dtype'] = fallback_dtype
                        yield op_cfgs
                        acc, _ = self.last_tune_result

                        if not isinstance(acc, list):
                            if ((self.higher_is_better and acc <= best_acc) \
                                or (not self.higher_is_better and acc >= best_acc)):
                                op_cfgs['op'][op] = copy.deepcopy(old_cfg)
                            else:
                                best_acc = acc
                        elif len(self.metric_name) > 1 and self.metric_weight is not None:
                            acc = np.mean(np.array(acc) * self.metric_weight)
                            if (self.higher_is_better and acc <= best_acc) or \
                                (not self.higher_is_better and acc >= best_acc):
                                op_cfgs['op'][op] = copy.deepcopy(old_cfg)
                            else:
                                best_acc = acc
                        elif len(self.metric_name) > 1 and self.metric_weight is None:
                            if all([acc_i >= best_i if higher_is_better else acc_i <= best_i for \
                                acc_i, best_i, higher_is_better in \
                                zip(acc, best_acc, self.metric_criterion)]):
                                op_cfgs['op'][op] = copy.deepcopy(old_cfg)
                            else:
                                best_acc = acc
 
                    op_cfgs = copy.deepcopy(best_cfg)
                    for op in ordered_ops:
                        for cfg in self.opwise_tune_cfgs[op]:
                            if ('weight' in cfg and fallback_dtype == cfg['weight']['dtype']) \
                               or ('weight' not in cfg
                                   and fallback_dtype == cfg['activation']['dtype']):
                                if op_cfgs['op'][op]['weight'] is not None:
                                    op_cfgs['op'][op]['weight'].clear()
                                    op_cfgs['op'][op]['weight']['dtype'] = fallback_dtype
                                assert cfg['activation']['dtype'] == fallback_dtype
                                op_cfgs['op'][op]['activation'].clear()
                                op_cfgs['op'][op]['activation']['dtype'] = fallback_dtype
                        yield op_cfgs
        else:
            logger.debug(self.opwise_tune_cfgs)
            op_cfgs['op'] = OrderedDict()
            for op in self.opwise_tune_cfgs.keys():
                op_cfgs['op'][op] = copy.deepcopy(self.opwise_tune_cfgs[op][0])
            yield op_cfgs

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .conf.config import DistillationConf, PruningConf, QuantConf
from .experimental.common import Model
from .experimental.distillation import Distillation
from .experimental.pruning import Pruning
from .experimental.quantization import Quantization
from .experimental.scheduler import Scheduler
from .model.base_model import BaseModel
from .utils import logger
from typing import Callable, List, Union


class TrainingCallbacks:
    """TrainingCallbacks is uesd in train loop for what user want to deal with additional
    arguments:
        commponent: one instance of Distillation, Quantization, Pruning, Scheduler

    examples:
        import neural_compressor.training.prepare
        training_callbacks = prepare(conf, model)
        train_loop:
            training_callbacks.on_train_begin()
            for epoch in range(epochs):
                training_callbacks.on_epoch_begin(epoch)
                for i, batch in enumerate(dataloader):
                    training_callbacks.on_step_begin(i)
                    ......
                    output = model(batch)
                    loss = ......
                    loss = training_callbacks.on_after_compute_loss(batch, output, loss)
                    loss.backward()
                    training_callbacks.on_before_optimizer_step()
                    optimizer.step()
                    training_callbacks.on_step_end()
                training_callbacks.on_epoch_end()
            training_callbacks.on_train_end()
    """
    def __init__(self, component):
        self.component = component
        self._component = \
            component.components[0] if isinstance(component, Scheduler) else component

    def on_train_begin(self):
        """ called before the beginning of training"""
        for on_train_begin_hook in self._component.hooks_dict['on_train_begin']:
            on_train_begin_hook()

    def on_train_end(self):
        """ called after the end of epochs"""
        for on_train_end_hook in self._component.hooks_dict['on_train_end']:
            on_train_end_hook()

    def on_epoch_begin(self, epoch):
        """ called on the beginning of epochs"""
        for on_epoch_begin_hook in self._component.hooks_dict['on_epoch_begin']:
            on_epoch_begin_hook(epoch)

    def on_step_begin(self, batch_id):
        """ called on the beginning of batches"""
        res_list = []
        for on_step_begin_hook in self._component.hooks_dict['on_step_begin']:
            res_list.append(on_step_begin_hook(batch_id))
        return res_list

    def on_after_compute_loss(self, input, student_output, student_loss, teacher_output=None):
        """ called on the end of loss computation"""
        loss = student_loss
        for on_after_compute_loss_hook in self._component.hooks_dict['on_after_compute_loss']:
            loss = on_after_compute_loss_hook(input, student_output, loss, teacher_output)
        return loss

    def on_before_optimizer_step(self):
        """ called on the end of backward"""
        for on_before_optimizer_step_hook in self._component.hooks_dict['on_before_optimizer_step']:
            on_before_optimizer_step_hook()

    def on_step_end(self):
        """ called on the end of batches"""
        res_list = []
        for on_step_end_hook in self._component.hooks_dict['on_step_end']:
            res_list.append(on_step_end_hook())
        return res_list

    def on_epoch_end(self):
        """ called on the end of epochs"""
        res_list = []

        for on_epoch_end_hook in self._component.hooks_dict['on_epoch_end']:
            res_list.append(on_epoch_end_hook())

        return res_list


def prepare(
    confs: Union[Callable, List], model: Callable = None, teacher_model: Callable = None,
    features=None
):
    """_summary_

    Args:
        confs (Union[Callable, List]): config of Distillation, Quantization, Pruning,
                                       or list of config for orchestration optimization
        model (Callable, optional): model to optimize. Defaults to None.
        teacher_model (Callable, optional): teacher model for distillation. Defaults to None.
        features (optional): teacher features for distillation, features and teacher_model are alternative.
                                     Defaults to None.

    Returns:
        TrainingCallbacks
        model

    examples:
        import neural_compressor.training.prepare
        training_callbacks = prepare(conf, model)
        train_loop:
            training_callbacks.on_train_begin()
            for epoch in range(epochs):
                training_callbacks.on_epoch_begin(epoch)
                for i, batch in enumerate(dataloader):
                    training_callbacks.on_step_begin(i)
                    ......
                    output = model(batch)
                    loss = ......
                    loss = training_callbacks.on_after_compute_loss(batch, output, loss)
                    loss.backward()
                    training_callbacks.on_before_optimizer_step()
                    optimizer.step()
                    training_callbacks.on_step_end()
                training_callbacks.on_epoch_end()
            training_callbacks.on_train_end()
    """
    if isinstance(confs, List):
        from .experimental.scheduler import Scheduler
        comps = []
        for config in confs:
            if isinstance(config, QuantConf):
                com = Quantization(config)
            elif isinstance(config, PruningConf):
                com = Pruning(config)
            elif isinstance(config, DistillationConf):
                com = Distillation(config)
                if teacher_model is not None:
                    com.teacher_model = teacher_model
            else:
                assert False, "Unsupported component: {}".format(config)

            comps.append(com)
        scheduler = Scheduler()
        assert model is not None, "need set neural_compressor Model for scheduler."
        scheduler.model = model
        comp = scheduler.combine(*comps)
        scheduler.append(comp)
        component = scheduler
    else:
        if isinstance(confs, QuantConf):
            component = Quantization(confs)
        elif isinstance(confs, PruningConf):
            component = Pruning(confs)
        elif isinstance(confs, DistillationConf):
            component = Distillation(confs)
            if teacher_model is not None:
                component.teacher_model = teacher_model
        else:
            assert False, logger.error(
                "confs should be one of QuantConf, PruningConf, DistillationConf. not ", confs
            )

        if model is not None:
            component.model = Model(model)

    return TrainingCallbacks(component), component.model


def fit(
    model: BaseModel,
    callbacks: TrainingCallbacks,
    train_func: Callable,
    eval_func: Callable
):
    """_summary_

    Args:
        model: neural_compressor.model.base_model.BaseModel, which to be optimized
        callbacks (TrainingCallbacks): instance of TrainingCallbacks which from prepare function
        train_func (function, optional): _description_. Defaults to None.
        eval_func (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    examples:
        import neural_compressor.training.prepare
        training_callbacks = prepare(conf, model)
        def train_func(model):
            training_callbacks.on_train_begin()
            for epoch in range(epochs):
                training_callbacks.on_epoch_begin(epoch)
                for i, batch in enumerate(dataloader):
                    training_callbacks.on_step_begin(i)
                    ......
                    output = model(batch)
                    loss = ......
                    loss = training_callbacks.on_after_compute_loss(batch, output, loss)
                    loss.backward()
                    training_callbacks.on_before_optimizer_step()
                    optimizer.step()
                    training_callbacks.on_step_end()
                training_callbacks.on_epoch_end()
            training_callbacks.on_train_end()
            return model

        def eval_func(model):
            ......

        fit(model, training_callbacks, train_func, eval_func)
    """
    callbacks.component.model = model
    assert train_func is not None, "train_func must be set!"
    assert eval_func is not None, "eval_func must be set!"
    if isinstance(callbacks.component, Scheduler):
        callbacks.component.components[0].train_func = train_func
    else:
        callbacks.component.train_func = train_func
    if isinstance(callbacks.component, Scheduler):
        callbacks.component.components[0].eval_func = eval_func
    else:
        callbacks.component.eval_func = eval_func

    return callbacks.component()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf

from tensorflow.python.framework import tensor_util
from tensorflow.core.framework import attr_value_pb2
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from ..graph_base import GraphRewriterBase
from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer
from neural_compressor.adaptor.tf_utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.adaptor.tf_utils.util import version1_gt_version2, version1_lt_version2


class FuseMatMulRequantizeDequantizeTransformer(GraphRewriterBase):
    """Fuse QuantizedMatMul + Requantize + Dequantize into QuantizedMatMulWithBiasAndDequantize.
    """
    def __init__(self, model, device='cpu'):
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

        self.eps = 1e-5

    def do_transformation(self):
        fuse_pattern = []
        if tf.version.VERSION in ("1.15.0-up2", "1.15.0-up3") or version1_gt_version2(tf.version.VERSION, '2.1.0'):
            fuse_pattern = [["QuantizedMatMulWithBias"], ['Requantize'], ['Dequantize'], ('Softmax',)]
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(fuse_pattern)
        for i in target_nodes:
            # TODO Remove below checker once the TF's limitation removed.
            if len(i) == 5 and version1_lt_version2(tf.__version__, '2.6.0'):
                continue

            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]
            deq_node_name = i[2]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndDequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if 'T1' in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr['T1'])
            if 'T2' in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr['T2'])

            top_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = self.graph_info[new_node.input[6]].node
            min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node

            bias_node = self.graph_info[Helper.node_name_from_input(new_node.input[2])].node
            max_input_node = self.graph_info[last_node.input[-1]].node
            min_input_node = self.graph_info[last_node.input[-2]].node

            if max_filter_node.op == 'Const':
                min_input_value = (min_input_node.attr['value'].tensor.float_val)[0]
                max_input_value = (max_input_node.attr['value'].tensor.float_val)[0]

                max_filter_value = (max_filter_node.attr['value'].tensor.float_val)[0]
                min_filter_value = (min_filter_node.attr['value'].tensor.float_val)[0]

                weights_tensor = tensor_util.MakeNdarray(
                        self.graph_info[new_node.input[1]].node.attr['value'].tensor)
                bias_tensor = tensor_util.MakeNdarray(
                    self.graph_info[new_node.input[2]].node.attr['value'].tensor)
                is_min_first = bool(quantized_node.attr['input_quant_mode'].s == b'MIN_FIRST')
                input_range = max_input_value - min_input_value if is_min_first else max(
                        abs(max_input_value), abs(min_input_value))

                if  -self.eps <= input_range <= self.eps:
                    input_range += self.eps

                if -self.eps <= max_input_value - min_input_value <= self.eps:
                    max_input_value += self.eps
                int32_bias = Helper.generate_int32_bias_for_matmul(bias_tensor, weights_tensor,
                                                        input_range, max_input_value,
                                                        min_input_value,
                                                        max_filter_value, min_filter_value)

                bias_node.attr['dtype'].CopyFrom(
                    attr_value_pb2.AttrValue(
                        type=float32_type if self.device == 'gpu' else qint32_type))
                bias_node.attr['value'].CopyFrom(
                    attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(
                        bias_tensor if self.device == 'gpu' else int32_bias, dtypes.
                        float32 if self.device == 'gpu' else dtypes.int32, bias_tensor.shape)))

                bias_node.attr['value'].tensor.dtype = float32_type \
                                        if self.device == 'gpu' else qint32_type
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type \
                                                if self.device == 'gpu' else qint32_type))
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            new_node.attr["Toutput"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            self.graph_analyzer.remove_node(requantize_node_name)

            if self.graph_info[deq_node_name].outputs:
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name,
                    self.graph_info[deq_node_name].outputs, deq_node_name)
                self.graph_analyzer.remove_node(deq_node_name)
            else:
                self.graph_analyzer.remove_node(deq_node_name)

                new_node.name = deq_node_name
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name, [], deq_node_name)

            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()
class FuseMatMulRequantizeTransformer(GraphRewriterBase):
    """Fuse Quantized MatMul Op with the successor Requantize Op.
    """
    def __init__(self, model, device='cpu'):
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.eps = 1e-05
        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.
        Returns:
            [graphdef]: the optimized graphdef object
        """
        uint8_type = dtypes.quint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum

        while True:
            target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(
                [["QuantizedMatMulWithBiasAndRelu"], ['Requantize']])
            if len(target_nodes) == 0:
                break

            i = target_nodes[0]
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)
            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if 'T1' in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr['T1'])
            if 'T2' in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr['T2'])

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = self.graph_info[new_node.input[6]].node
            min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node

            is_min_first = bool(quantized_node.attr['input_quant_mode'].s == b'MIN_FIRST')

            bias_node = self.graph_info[new_node.input[2]].node
            max_input_node = self.graph_info[last_node.input[-1]].node
            min_input_node = self.graph_info[last_node.input[-2]].node

            if last_node.op.find('Requantize') != -1 or last_node.op.find('QuantizeV2') != -1:
                min_input_value = (min_input_node.attr['value'].tensor.float_val)[0]
                max_input_value = (max_input_node.attr['value'].tensor.float_val)[0]

                max_filter_value = (max_filter_node.attr['value'].tensor.float_val)[0]
                min_filter_value = (min_filter_node.attr['value'].tensor.float_val)[0]

                weights_tensor = tensor_util.MakeNdarray(
                        self.graph_info[new_node.input[1]].node.attr['value'].tensor)
                bias_tensor = tensor_util.MakeNdarray(
                    self.graph_info[new_node.input[2]].node.attr['value'].tensor)

                input_range = max_input_value - min_input_value if is_min_first else max(
                        abs(max_input_value), abs(min_input_value))

                if -self.eps <= input_range <= self.eps:
                    input_range += self.eps

                if -self.eps <= max_input_value - min_input_value <= self.eps:
                    max_input_value += self.eps
                int32_bias = Helper.generate_int32_bias_for_matmul(bias_tensor, weights_tensor,
                                                                   input_range,
                                                                   max_input_value,
                                                                   min_input_value,
                                                                   max_filter_value,
                                                                   min_filter_value)
                bias_node.attr['dtype'].CopyFrom(
                    attr_value_pb2.AttrValue(
                        type=float32_type if self.device == 'gpu' else qint32_type))
                bias_node.attr['value'].CopyFrom(
                    attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(
                        bias_tensor if self.device == 'gpu' else int32_bias, dtypes.
                        float32 if self.device == 'gpu' else dtypes.int32, bias_tensor.shape)))

                bias_node.attr['value'].tensor.dtype = float32_type \
                                        if self.device == 'gpu' else qint32_type
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type \
                                                if self.device == 'gpu' else qint32_type))

                new_node.attr["Toutput"].CopyFrom(
                        attr_value_pb2.AttrValue(type=uint8_type))
                #TODO enabled below commit once the graph refactor pre_optimize commmitted.
                if quantized_node_op.find('Relu') == -1:
                    deq_node_name = self.graph_info[requantize_node_name].outputs[0]
                    deq_node = self.graph_info[deq_node_name].node
                    deq_node.attr['T'].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            self.graph_analyzer.replace_single_node(
                new_node, [parent_node_name], quantized_node_name,
                [self.graph_info[requantize_node_name].outputs[0]], requantize_node_name)
            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()

class FuseMatMulRequantizeDequantizeNewAPITransformer(GraphRewriterBase): # pragma: no cover
    """Fuse _QuantizedFusedMatMul + Requantize + Dequantize into _QuantizedFusedMatMulAndDequantize.
    """
    def __init__(self, model, device='cpu'):
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

        self.eps = 1e-5

    def do_transformation(self):
        fuse_pattern = [["_QuantizedFusedMatMul"], ['Requantize'], ['Dequantize'], ('Softmax',)]

        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(fuse_pattern)
        for i in target_nodes:
            # TODO Remove below checker once the TF's limitation removed.
            if len(i) == 5:
                continue

            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]
            deq_node_name = i[2]

            quantized_node_op = i[-1][0]
            
            if "BiasAdd" in quantized_node.attr["fused_ops"].SerializeToString().decode('UTF-8') \
               and "Relu" in quantized_node.attr["fused_ops"].SerializeToString().decode('UTF-8'):
                continue

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndDequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            #new_node.input.append(requested_output_min_name)
            #new_node.input.append(requested_output_max_name)
            if 'T1' in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr['T1'])
            if 'T2' in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr['T2'])
            if 'num_args' in quantized_node.attr:
                new_node.attr["num_args"].CopyFrom(quantized_node.attr['num_args'])
            if 'fused_ops' in quantized_node.attr:
                new_node.attr["fused_ops"].CopyFrom(quantized_node.attr["fused_ops"])

            top_node_name = Helper.node_name_from_input(quantized_node.input[0])
            float32_type = dtypes.float32.as_datatype_enum
            new_node.attr["Targs"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            new_node.attr["Toutput"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))

            self.graph_analyzer.remove_node(requantize_node_name)

            if self.graph_info[deq_node_name].outputs:
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name,
                    self.graph_info[deq_node_name].outputs, deq_node_name)
                self.graph_analyzer.remove_node(deq_node_name)
            else:
                self.graph_analyzer.remove_node(deq_node_name)

                new_node.name = deq_node_name
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name, [], deq_node_name)

            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()

class FuseMatMulRequantizeNewAPITransformer(GraphRewriterBase): # pragma: no cover
    """Fuse newAPI Quantized MatMul Op with the successor Requantize Op.
    """
    def __init__(self, model, device='cpu'):
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.eps = 1e-05
        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.
        Returns:
            [graphdef]: the optimized graphdef object
        """
        uint8_type = dtypes.quint8.as_datatype_enum

        while True:
            target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(
                [["_QuantizedFusedMatMul"], ['Requantize']])
            if len(target_nodes) == 0:
                break

            i = target_nodes[0]
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]
            if "BiasAdd" in quantized_node.attr["fused_ops"].SerializeToString().decode('UTF-8') \
               and "Relu" not in quantized_node.attr["fused_ops"].SerializeToString().decode('UTF-8'):
                break

            new_node = node_def_pb2.NodeDef()

            new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)
            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if 'T1' in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr['T1'])
            if 'T2' in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr['T2'])
            if 'num_args' in quantized_node.attr:
                new_node.attr["num_args"].CopyFrom(quantized_node.attr["num_args"])
            if 'Targs' in quantized_node.attr:
                new_node.attr["Targs"].CopyFrom(quantized_node.attr["Targs"])
            if 'fused_ops' in quantized_node.attr:
                new_node.attr["fused_ops"].CopyFrom(quantized_node.attr["fused_ops"])
            new_node.attr["Toutput"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            self.graph_analyzer.replace_single_node(
                new_node, [parent_node_name], quantized_node_name,
                [self.graph_info[requantize_node_name].outputs[0]], requantize_node_name)
            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from .direct_q8 import QDQDirect8BitOp
from onnxruntime.quantization.quant_utils import QuantizedValueType
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue


class QMaxPool(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        assert (node.op_type == "MaxPool")

        if self.quantizer.opset_version < 12: # pragma: no cover
            return

        if len(self.quantizer.model.get_children(node)) == 0:
            return
        parent = self.quantizer.model.get_parents(node)[0]
        children = self.quantizer.model.get_children(node)
        if parent.op_type != 'DequantizeLinear' or \
            all([i.op_type != 'QuantizeLinear' for i in children]):
            return
        node.input[0] = parent.input[0]
        node.output[0] = node.output[0] + '_quantized'
        for child in children:
            if child.op_type == 'QuantizeLinear':
                self.quantizer.remove_nodes.append(child)
                for n in self.quantizer.model.get_children(child):
                    self.quantizer.model.replace_node_input(n,
                                    child.output[0], node.output[0])

        self.quantizer.remove_nodes.append(parent)

class QDQMaxPool(QDQDirect8BitOp):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert (node.op_type == "MaxPool")

        # if version is less than 12, just no change
        if self.quantizer.opset_version < 12:
            return

        # Direct 8bits op
        super().quantize()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neural_compressor.utils.utility import dump_elapsed_time

from ..graph_base import GraphRewriterBase
from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer
from neural_compressor.adaptor.tf_utils.graph_util import GraphRewriterHelper as Helper


class MetaInfoChangingMemOpOptimizer(GraphRewriterBase):
    """Fuse the pattern like Dequantize + MetaOp + Quantize into MetaOp(set its type to int8).
       With such changes, the Quantize and Dequantize OP will removed for better performance.
    """
    meta_op_type_list = ['Reshape', 'Squeeze']

    def __init__(self, model):
        super().__init__(model)

        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

    def _check_and_apply_transform(self, quantize_node_name):
        cur_node = self.graph_info[quantize_node_name].node
        res = [quantize_node_name]
        found_meta_op_flag = False
        while cur_node.input:
            pre_node_name = Helper.node_name_from_input(cur_node.input[0])
            pre_node = self.graph_info[pre_node_name].node
            pre_node_op = pre_node.op
            if pre_node_op in self.meta_op_type_list:
                res.insert(0, pre_node_name)
                cur_node = pre_node
            elif pre_node_op == 'Dequantize' and len(self.graph_info[pre_node_name].outputs) == 1:
                res.insert(0, pre_node_name)
                found_meta_op_flag = True
                break
            else:
                break

        dequantize_node_name = res[0]
        quantize_node_name = res[-1]
        deq_node = self.graph_info[dequantize_node_name].node
        quant_node = self.graph_info[quantize_node_name].node

        if found_meta_op_flag and len(res) > 2 and \
            quant_node.attr['mode'].s.decode() == deq_node.attr['mode'].s.decode():
            deq_min_range = self.graph_info[dequantize_node_name].node.input[1]
            deq_max_range = self.graph_info[dequantize_node_name].node.input[2]
            quant_output_min = quantize_node_name + ':1'
            quant_output_max = quantize_node_name + ':2'
            quantize_input_name = res[-2]

            quantized_node_name = self.graph_info[quantize_node_name].outputs[0]

            for index, value in enumerate(self.graph_info[quantized_node_name].node.input):
                if value == quant_output_min:
                    self.graph_info[quantized_node_name].node.input[index] = deq_min_range

                if value == quant_output_max:
                    self.graph_info[quantized_node_name].node.input[index] = deq_max_range

                if index == 0:
                    self.graph_info[quantized_node_name].node.input[index] = quantize_input_name

            new_dtype = self.graph_info[dequantize_node_name].node.attr['T'].type
            for node_name in res[1: -1]:
                self.graph_info[node_name].node.attr['T'].type = new_dtype

            if 'T1' in self.graph_info[quantized_node_name].node.attr:
                self.graph_info[quantized_node_name].node.attr['T1'].type = new_dtype

            if 'Tinput' in self.graph_info[quantized_node_name].node.attr:
                self.graph_info[quantized_node_name].node.attr['Tinput'].type = new_dtype

            if 'T' in self.graph_info[quantized_node_name].node.attr:
                self.graph_info[quantized_node_name].node.attr['T'].type = new_dtype

            self.graph_info[res[1]
                            ].node.input[0] = self.graph_info[dequantize_node_name].node.input[0]
            self.graph_analyzer.remove_node(dequantize_node_name)
            self.graph_analyzer.remove_node(self.graph_info[quantize_node_name].node.input[1])
            self.graph_analyzer.remove_node(self.graph_info[quantize_node_name].node.input[2])
            self.graph_analyzer.remove_node(quantize_node_name)

    @dump_elapsed_time("Pass MetaOpOptimizer")
    def do_transformation(self):

        deq_node = self.graph_analyzer.query_fusion_pattern_nodes([['QuantizeV2']])

        for i in deq_node:
            self._check_and_apply_transform(i[0])
        return GraphAnalyzer().dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
import os
from neural_compressor.utils.utility import LazyImport, singleton
from PIL import Image
torch = LazyImport('torch')
torchvision = LazyImport('torchvision')
tf = LazyImport('tensorflow')
mx = LazyImport('mxnet')
np = LazyImport('numpy')
hashlib = LazyImport('hashlib')
gzip = LazyImport('gzip')
tarfile = LazyImport('tarfile')
zipfile = LazyImport('zipfile')
pickle = LazyImport('pickle')
glob = LazyImport('glob')

@singleton
class TensorflowDatasets(object):
    def __init__(self):
        self.datasets = {}
        self.datasets.update(TENSORFLOW_DATASETS)


@singleton
class PyTorchDatasets(object):
    def __init__(self):
        self.datasets = {
            'ImageFolder': PytorchMxnetWrapDataset(
                    torchvision.datasets.ImageFolder),
        }
        self.datasets.update(PYTORCH_DATASETS)

@singleton
class MXNetDatasets(object):
    def __init__(self):
        self.datasets = {}
        self.datasets.update(MXNET_DATASETS)

@singleton
class ONNXRTQLDatasets(object):
    def __init__(self):
        self.datasets = {}
        self.datasets.update(ONNXRTQL_DATASETS)

@singleton
class ONNXRTITDatasets(object):
    def __init__(self):
        self.datasets = {}
        self.datasets.update(ONNXRTIT_DATASETS)

class PytorchMxnetWrapDataset():
    def __init__(self, datafunc):
        self.datafunc = datafunc

    def __call__(self, transform=None, filter=None, *args, **kwargs):
        return PytorchMxnetWrapFunction(self.datafunc, transform=transform, \
                        filter=filter, *args, **kwargs)

class PytorchMxnetWrapFunction():
    def __init__(self, dataset, transform, filter, *args, **kwargs):
        self.dataset = dataset(*args, **kwargs)
        self.transform = transform
        self.filter = filter

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        sample = self.dataset[index]
        if self.transform is not None:
            sample = self.transform(sample)
        return sample


framework_datasets = {"tensorflow": TensorflowDatasets,
                      "inteltensorflow": TensorflowDatasets,
                      "tensorflow_itex": TensorflowDatasets,
                      "mxnet": MXNetDatasets,
                      "pytorch": PyTorchDatasets,
                      "pytorch_ipex": PyTorchDatasets,
                      "pytorch_fx": PyTorchDatasets,
                      "onnxrt_qdq": ONNXRTQLDatasets,
                      "onnxrt_qlinearops": ONNXRTQLDatasets,
                      "onnxrt_qoperator": ONNXRTQLDatasets,
                      "onnxrt_integerops": ONNXRTITDatasets,
                      }

"""The datasets supported by neural_compressor, it's model specific and can be configured by yaml file.

   User could add new datasets by implementing new Dataset subclass under this directory.
   The naming convention of new dataset subclass should be something like ImageClassifier, user
   could choose this dataset by setting "imageclassifier" string in tuning.strategy field of yaml.

   DATASETS variable is used to store all implemented Dataset subclasses to support
   model specific dataset.
"""


class DATASETS(object):
    def __init__(self, framework):
        assert framework in ["inteltensorflow", "tensorflow", "tensorflow_itex", \
                             "mxnet", "onnxrt_qdq", "onnxrt_qlinearops", "onnxrt_integerops", \
                             "pytorch", "pytorch_ipex", "pytorch_fx", "onnxrt_qoperator"], \
                             "framework support tensorflow pytorch mxnet onnxrt"
        self.datasets = framework_datasets[framework]().datasets

    def __getitem__(self, dataset_type):
        assert dataset_type in self.datasets.keys(), "dataset type only support {}".\
            format(self.datasets.keys())
        return self.datasets[dataset_type]

# user/model specific datasets will be registered here
TENSORFLOW_DATASETS = {}
INTELTENSORFLOW_DATASETS = {}
TENSORFLOWITEX_DATASETS = {}
MXNET_DATASETS = {}
PYTORCH_DATASETS = {}
PYTORCHIPEX_DATASETS = {}
PYTORCHFX_DATASETS = {}
ONNXRTQL_DATASETS = {}
ONNXRTIT_DATASETS = {}

registry_datasets = {"tensorflow": TENSORFLOW_DATASETS,
                     "inteltensorflow": INTELTENSORFLOW_DATASETS,
                     "tensorflow_itex": TENSORFLOWITEX_DATASETS,
                     "mxnet": MXNET_DATASETS,
                     "pytorch": PYTORCH_DATASETS,
                     "pytorch_ipex": PYTORCHIPEX_DATASETS,
                     "pytorch_fx": PYTORCHFX_DATASETS,
                     "onnxrt_integerops": ONNXRTIT_DATASETS,
                     "onnxrt_qdq": ONNXRTQL_DATASETS,
                     "onnxrt_qoperator": ONNXRTQL_DATASETS,
                     "onnxrt_qlinearops": ONNXRTQL_DATASETS,
                    }


def dataset_registry(dataset_type, framework, dataset_format=''):
    """The class decorator used to register all Dataset subclasses.


    Args:
        cls (class): The class of register.
        dataset_type (str): The dataset registration name
        framework (str): support 3 framework including 'tensorflow', 'pytorch', 'mxnet'
        data_format (str): The format dataset saved, eg 'raw_image', 'tfrecord'

    Returns:
        cls: The class of register.
    """
    def decorator_dataset(cls):
        for single_framework in [fwk.strip() for fwk in framework.split(',')]:
            assert single_framework in [
                "tensorflow",
                "inteltensorflow",
                "tensorflow_itex",
                "mxnet",
                "pytorch",
                "pytorch_ipex",
                "pytorch_fx",
                "onnxrt_qlinearops",
                "onnxrt_integerops",
                "onnxrt_qdq",
                "onnxrt_qoperator",
            ], "The framework support tensorflow mxnet pytorch onnxrt"
            dataset_name = dataset_type + dataset_format
            if dataset_name in registry_datasets[single_framework].keys():
                raise ValueError('Cannot have two datasets with the same name')
            registry_datasets[single_framework][dataset_name] = cls
        return cls
    return decorator_dataset


class Dataset(object):
    """ The base class of dataset. Subclass datasets should overwrite two methods:
    `__getitem__` for indexing to data sample and `__len__`for the size of the dataset

    """

    @abstractmethod
    def __getitem__(self, index):
        raise NotImplementedError

    # it's suggested to implement your  __len__ method though we do not set it in abstract class
    # @abstractmethod
    # def __len__(self):
    #     raise NotImplementedError

class IterableDataset(object):
    """An iterable Dataset. Subclass iterable dataset should also implement a method:
    `__iter__` for interating over the samples of the dataset.

    """

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError

def download_url(url, root, filename=None, md5=None): # pragma: no cover
    import urllib
    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    if check_integrity(fpath, md5):
        print('Using downloaded and verified file: ' + fpath)
    else:
        try:
            print('Downloading ' + url + ' to ' + fpath)
            urllib.request.urlretrieve(
                url, fpath,
                reporthook=gen_bar_updater()
            )
        except (urllib.error.URLError, IOError) as e:
            if url[:5] == 'https':
                url = url.replace('https:', 'http:')
                print('Failed download. Trying https -> http instead.'
                      ' Downloading ' + url + ' to ' + fpath)
                urllib.request.urlretrieve(
                    url, fpath,
                    reporthook=gen_bar_updater()
                )
            else:
                raise e
        if not check_integrity(fpath, md5):
            raise RuntimeError("File not found or corrupted.")

def gen_bar_updater():
    from tqdm import tqdm
    pbar = tqdm(total=None)
    def bar_update(count, block_size, total_size):
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = count * block_size
        pbar.update(progress_bytes - pbar.n)
    return bar_update

def check_integrity(fpath, md5):
    if not os.path.isfile(fpath):
        return False
    if md5 is None:
        return True
    return md5 == calculate_md5(fpath)

def calculate_md5(fpath, chunk_size=1024*1024):
    md5 = hashlib.md5()
    with open(fpath, 'rb') as f:
        for chunk in iter(lambda: f.read(chunk_size), b''):
            md5.update(chunk)
    return md5.hexdigest()

@dataset_registry(dataset_type="CIFAR10", framework="onnxrt_qlinearops, \
                    onnxrt_integerops", dataset_format='')
class CIFAR10(Dataset):
    """Configuration for CIFAR10 and CIFAR100 database

    For CIFAR10: If download is True, it will download dataset to root/ and extract it
                 automatically, otherwise user can download file from
                 https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz manually to
                 root/ and extract it.
    For CIFAR100: If download is True, it will download dataset to root/ and extract it
                  automatically, otherwise user can download file from
                  https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz manually to
                  root/ and extract it.

    Args:
        root (str): Root directory of dataset.
        train (bool, default=False): If True, creates dataset from train subset,
                                     otherwise from validation subset.
        transform (transform object, default=None):  transform to process input data.
        filter (Filter objects, default=None): filter out examples according to specific
                                               conditions.
        download (bool, default=True): If true, downloads the dataset from the internet
                                       and puts it in root directory. If dataset is already
                                       downloaded, it is not downloaded again.
    """
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    filename = "cifar-10-python.tar.gz"
    tgz_md5 = 'c58f30108f718f92721af3b95e74349a'
    train_list = [
        ['data_batch_1', 'c99cafc152244af753f735de768cd75f'],
        ['data_batch_2', 'd4bba439e000b95fd0a9bffe97cbabec'],
        ['data_batch_3', '54ebc095f3ab1f0389bbae665268c751'],
        ['data_batch_4', '634d18415352ddfa80567beed471001a'],
        ['data_batch_5', '482c414d41f54cd18b22e5b47cb7c3cb'],
    ]

    test_list = [
        ['test_batch', '40351d587109b95175f43aff81a1287e'],
    ]

    meta = {
        'filename': 'batches.meta',
        'key': 'label_names',
        'md5': '5ff9c542aee3614f3951f8cda6e48888',
    }

    def __init__(self, root, train=False, transform=None, filter=None,
                    download=True): # pragma: no cover
        self.root = root
        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                'Dataset not found or corrupted. You can use download=True to download it')
        if train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data = []
        self.targets = []
        for file_name, checksum in downloaded_list:
            file_path = os.path.join(self.root, file_name)
            with open(file_path, 'rb') as f:
                entry = pickle.load(f, encoding='latin1')
                self.data.append(entry['data'])
                if 'labels' in entry:
                    self.targets.extend(entry['labels'])
                else:
                    self.targets.extend(entry['fine_labels'])
        self.data = np.vstack(self.data).reshape(-1, 3, 32, 32)
        self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC

        self.load_meta()
        self.transform = transform

    def load_meta(self): # pragma: no cover
        path = os.path.join(self.root, self.meta['filename'])
        if not check_integrity(path, self.meta['md5']):
            raise RuntimeError('Dataset metadata file not found or corrupted.' +
                               ' You can use download=True to download it')
        with open(path, 'rb') as infile:
            data = pickle.load(infile, encoding='latin1')
            self.classes = data[self.meta['key']]
        self.class_to_idx = {_class: i for i, _class in enumerate(self.classes)}

    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        if self.transform is not None:
            image, label = self.transform((image, label))
        return image, label

    def __len__(self): # pragma: no cover
        return len(self.data)

    def download(self): # pragma: no cover
        if self._check_integrity():
            print('Files already downloaded and verified')
            return
        download_root = os.path.expanduser(self.root)
        filename = os.path.basename(self.url)
        download_url(self.url, download_root, filename, self.tgz_md5)
        archive = os.path.join(download_root, filename)
        print("Extracting {} to {}".format(archive, download_root))
        with tarfile.open(archive, 'r:gz') as tar:
            tar.extractall(path=download_root)

    def _check_integrity(self): # pragma: no cover
        root = self.root
        for fentry in (self.train_list + self.test_list):
            filename, md5 = fentry[0], fentry[1]
            fpath = os.path.join(root, filename)
            if not check_integrity(fpath, md5):
                return False
            return True

@dataset_registry(dataset_type="CIFAR10", framework="pytorch", dataset_format='')
class PytorchCIFAR10(CIFAR10):
    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        image = Image.fromarray(image)
        if self.transform is not None:
            image, label = self.transform((image, label))
        return (image, label)

@dataset_registry(dataset_type="CIFAR10", framework="mxnet", dataset_format='')
class MXNetCIFAR10(CIFAR10):
    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        image = mx.nd.array(image)
        if self.transform is not None:
            image, label = self.transform((image, label))
        return (image, label)

@dataset_registry(dataset_type="CIFAR10", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowCIFAR10(CIFAR10):
    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        if self.transform is not None:
            image, label = self.transform((image, label))
        if type(image).__name__ == 'Tensor':
            with tf.compat.v1.Session() as sess:
                image = sess.run(image)
        elif type(image).__name__ == 'EagerTensor':
            image = image.numpy()
        return (image, label)

@dataset_registry(dataset_type="CIFAR100", framework="onnxrt_qlinearops, \
                    onnxrt_integerops", dataset_format='')
class CIFAR100(CIFAR10):
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = 'eb9058c3a382ffc7106e4002c42a8d85'
    train_list = [
        ['train', '16019d7e3df5f24257cddd939b257f8d'],
    ]
    test_list = [
        ['test', 'f0ef6b0ae62326f3e7ffdfab6717acfc'],
    ]
    meta = {
        'filename': 'meta',
        'key': 'fine_label_names',
        'md5': '7973b15100ade9c7d40fb424638fde48',
    }

@dataset_registry(dataset_type="CIFAR100", framework="pytorch", dataset_format='')
class PytorchCIFAR100(CIFAR100):
    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        image = Image.fromarray(image)
        if self.transform is not None:
            image, label = self.transform((image, label))
        image = np.array(image)
        return (image, label)

@dataset_registry(dataset_type="CIFAR100", framework="mxnet", dataset_format='')
class MXNetCIFAR100(CIFAR100):
    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        image = mx.nd.array(image)
        if self.transform is not None:
            image, label = self.transform((image, label))
        return (image, label)

@dataset_registry(dataset_type="CIFAR100", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowCIFAR100(CIFAR100):
    def __getitem__(self, index): # pragma: no cover
        image, label = self.data[index], self.targets[index]
        if self.transform is not None:
            image, label = self.transform((image, label))
        if type(image).__name__ == 'Tensor':
            with tf.compat.v1.Session() as sess:
                image = sess.run(image)
        elif type(image).__name__ == 'EagerTensor':
            image = image.numpy()
        return (image, label)

@dataset_registry(dataset_type="MNIST", framework="onnxrt_qlinearops, \
                    onnxrt_integerops", dataset_format='')
class MNIST(Dataset):
    """Configuration for Modified National Institute of Standards and Technology database
       and FashionMNIST database

    For MNIST: If download is True, it will download dataset to root/MNIST/, otherwise user
               should put mnist.npz under root/MNIST/ manually.
    For FashionMNIST: If download is True, it will download dataset to root/FashionMNIST/,
                      otherwise user should put train-labels-idx1-ubyte.gz,
                      train-images-idx3-ubyte.gz, t10k-labels-idx1-ubyte.gz
                      and t10k-images-idx3-ubyte.gz under root/FashionMNIST/ manually.

    Args:
        root (str): Root directory of dataset.
        train (bool, default=False): If True, creates dataset from train subset,
                                     otherwise from validation subset.
        transform (transform object, default=None):  transform to process input data.
        filter (Filter objects, default=None): filter out examples according to specific
                                               conditions.
        download (bool, default=True): If true, downloads the dataset from the internet
                                       and puts it in root directory. If dataset is already
                                       downloaded, it is not downloaded again.
    """
    classes = ['0 - zero', '1 - one', '2 - two', '3 - three', '4 - four',
                   '5 - five', '6 - six', '7 - seven', '8 - eight', '9 - nine']
    resource = [
        ('https://storage.googleapis.com/tensorflow/tf-keras-datasets/mnist.npz',
            '8a61469f7ea1b51cbae51d4f78837e45')
    ]

    def __init__(self, root, train=False, transform=None, filter=None, download=True):
        self.root = root
        self.train = train
        self.transform = transform
        if download:
            self.download()

        self.read_data()

    def read_data(self):
        for file_name, checksum in self.resource:
            file_path = os.path.join(self.root, os.path.basename(file_name))
            if not os.path.exists(file_path):
                raise RuntimeError(
                    'Dataset not found. You can use download=True to download it')
            with np.load(file_path, allow_pickle=True) as f:
                if self.train:
                    self.data, self.targets = f['x_train'], f['y_train']
                else:
                    self.data, self.targets = f['x_test'], f['y_test']

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = np.expand_dims(image, -1)
        if self.transform is not None:
            image, label = self.transform((image, label))
        return image, label

    @property
    def class_to_idx(self):
        return {_class: i for i, _class in enumerate(self.classes)}

    def download(self):
        for url, md5 in self.resource:
            filename = os.path.basename(url)
            if os.path.exists(os.path.join(self.root, filename)):
                continue
            else:
                download_url(url, root=self.root,
                            filename=filename, md5=md5)

@dataset_registry(dataset_type="MNIST", framework="pytorch", dataset_format='')
class PytorchMNIST(MNIST):
    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = Image.fromarray(image, mode='L')
        if self.transform is not None:
            image, label = self.transform((image, label))
        image = np.array(image)
        return (image, label)

@dataset_registry(dataset_type="MNIST", framework="mxnet", dataset_format='')
class MXNetMNIST(MNIST):
    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = mx.nd.array(image)
        image = image.reshape((image.shape[0], image.shape[1], 1))
        if self.transform is not None:
            image, label = self.transform((image, label))
        return (image, label)

@dataset_registry(dataset_type="MNIST", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowMNIST(MNIST):
    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = np.expand_dims(image, -1)
        if self.transform is not None:
            image, label = self.transform((image, label))
        if type(image).__name__ == 'Tensor':
            with tf.compat.v1.Session() as sess:
                image = sess.run(image)
        elif type(image).__name__ == 'EagerTensor':
            image = image.numpy()
        return (image, label)

@dataset_registry(dataset_type="FashionMNIST", framework="onnxrt_qlinearops, \
                    onnxrt_integerops", dataset_format='')
class FashionMNIST(MNIST):
    resource = [
        ('https://storage.googleapis.com/tensorflow/tf-keras-datasets/' + file_name, None)
        for file_name in [
            'train-labels-idx1-ubyte.gz', 'train-images-idx3-ubyte.gz',
            't10k-labels-idx1-ubyte.gz', 't10k-images-idx3-ubyte.gz'
            ]
    ]

    classes = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat', 'Sandal',
               'Shirt', 'Sneaker', 'Bag', 'Ankle boot']

    def read_data(self):
        import struct
        if self.train:
            label_path = os.path.join(self.root, 'train-labels-idx1-ubyte.gz')
            image_path = os.path.join(self.root, 'train-images-idx3-ubyte.gz')
        else:
            label_path = os.path.join(self.root, 't10k-labels-idx1-ubyte.gz')
            image_path = os.path.join(self.root, 't10k-images-idx3-ubyte.gz')
        with gzip.open(label_path, 'rb') as f:
            struct.unpack(">II", f.read(8))
            self.targets = np.frombuffer(f.read(), dtype=np.uint8).astype(np.int32)
        with gzip.open(image_path, 'rb') as f:
            struct.unpack(">IIII", f.read(16))
            data = np.frombuffer(f.read(), dtype=np.uint8)
            self.data = data.reshape(len(self.targets), 28, 28)

@dataset_registry(dataset_type="FashionMNIST", framework="pytorch", dataset_format='')
class PytorchFashionMNIST(FashionMNIST):
    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = Image.fromarray(image, mode='L')
        if self.transform is not None:
            image, label = self.transform((image, label))
        image = np.array(image)
        return (image, label)

@dataset_registry(dataset_type="FashionMNIST", framework="mxnet", dataset_format='')
class MXNetFashionMNIST(FashionMNIST):
    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = mx.nd.array(image)
        image = image.reshape((image.shape[0], image.shape[1], 1))
        if self.transform is not None:
            image, label = self.transform((image, label))
        return (image, label)

@dataset_registry(dataset_type="FashionMNIST", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowFashionMNIST(FashionMNIST):
    def __getitem__(self, index):
        image, label = self.data[index], int(self.targets[index])
        image = np.expand_dims(image, -1)
        if self.transform is not None:
            image, label = self.transform((image, label))
        if type(image).__name__ == 'Tensor':
            with tf.compat.v1.Session() as sess:
                image = sess.run(image)
        elif type(image).__name__ == 'EagerTensor':
            image = image.numpy()
        return (image, label)

@dataset_registry(dataset_type="ImageFolder", framework="onnxrt_qlinearops, \
                    onnxrt_integerops", dataset_format='')
class ImageFolder(Dataset):
    """Configuration for ImageFolder

    Expects the data folder to contain subfolders representing the classes to which
    its images belong.

    Please arrange data in this way:
        root/class_1/xxx.png
        root/class_1/xxy.png
        root/class_1/xxz.png
        ...
        root/class_n/123.png
        root/class_n/nsdf3.png
        root/class_n/asd932_.png
    Please put images of different categories into different folders.

    Args: root (str): Root directory of dataset.
          transform (transform object, default=None):  transform to process input data.
          filter (Filter objects, default=None): filter out examples according to specific
                                                 conditions
    """
    def __init__(self, root, transform=None, filter=None):
        self.root = root
        assert os.path.exists(self.root), "Datapath doesn't exist!"

        self.transform = transform
        self.image_list = []
        files = glob.glob(os.path.join(self.root, '*'))
        files.sort()
        for idx, file in enumerate(files):
            imgs = glob.glob(os.path.join(file, '*'))
            imgs.sort()
            for img in imgs:
                self.image_list.append((img, idx))

    def __len__(self):
        return len(self.image_list)

    def __getitem__(self, index):
        sample = self.image_list[index]
        label = sample[1]
        with Image.open(sample[0]) as image:
            image = np.array(image)
            if self.transform is not None:
                image, label = self.transform((image, label))
            return (image, label)

@dataset_registry(dataset_type="ImageFolder", framework="mxnet", dataset_format='')
class MXNetImageFolder(ImageFolder):
    def __getitem__(self, index):
        sample = self.image_list[index]
        label = sample[1]
        image = mx.image.imread(sample[0])
        if self.transform is not None:
            image, label = self.transform((image, label))
        return (image, label)

@dataset_registry(dataset_type="ImageFolder", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowImageFolder(ImageFolder):
    def __getitem__(self, index):
        sample = self.image_list[index]
        label = sample[1]
        with Image.open(sample[0]) as image:
            if image.mode != 'RGB':
                image = image.convert('RGB')
            image = np.array(image)
            if self.transform is not None:
                image, label = self.transform((image, label))
            if type(image).__name__ == 'Tensor':
                with tf.compat.v1.Session() as sess:
                    image = sess.run(image)
            elif type(image).__name__ == 'EagerTensor':
                image = image.numpy()
            return (image, label)

@dataset_registry(dataset_type="TFRecordDataset", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowTFRecordDataset(IterableDataset):
    """Configuration for TensorflowTFRecordDataset

    Root is a full path to tfrecord file, which contains the file name.

    Args: root (str): filename of dataset.
          transform (transform object, default=None):  transform to process input data.
          filter (Filter objects, default=None): filter out examples according
                                                 to specific conditions.
    """
    def __new__(cls, root, transform=None, filter=None):
        # pylint: disable=no-name-in-module
        from tensorflow.python.data.experimental import parallel_interleave
        from tensorflow.python.platform import gfile
        file_names = gfile.Glob(root)
        ds = tf.data.Dataset.from_tensor_slices(file_names)
        ds = ds.apply(parallel_interleave(
                tf.data.TFRecordDataset, cycle_length=len(file_names)))
        if transform is not None:
            ds = ds.map(transform, num_parallel_calls=None)
        ds = ds.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)  # this number can be tuned
        return ds

@dataset_registry(dataset_type="ImageRecord", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowImageRecord(IterableDataset):
    """Configuration for ImageNet database in tf record format

    Please arrange data in this way:
        root/validation-000-of-100
        root/validation-001-of-100
        ...
        root/validation-099-of-100
    The file name needs to follow this pattern: '* - * -of- *'

    Args: root (str): Root directory of dataset.
          transform (transform object, default=None):  transform to process input data.
          filter (Filter objects, default=None): filter out examples according
                                                 to specific conditions
    """

    """Configuration for Imagenet dataset."""
    def __new__(cls, root, transform=None, filter=None):

        from tensorflow.python.platform import gfile # pylint: disable=no-name-in-module
        glob_pattern = os.path.join(root, '*-*-of-*')
        file_names = gfile.Glob(glob_pattern)
        if not file_names:
            raise ValueError('Found no files in --root matching: {}'.format(glob_pattern))

        # pylint: disable=no-name-in-module
        from tensorflow.python.data.experimental import parallel_interleave
        from neural_compressor.experimental.data.transforms.imagenet_transform import ParseDecodeImagenet
        ds = tf.data.TFRecordDataset.list_files(file_names, shuffle=False)
        ds = ds.apply(parallel_interleave(
                tf.data.TFRecordDataset, cycle_length=len(file_names)))

        if transform is not None:
            transform.transform_list.insert(0, ParseDecodeImagenet())
        else:
            transform = ParseDecodeImagenet()
        ds = ds.map(transform, num_parallel_calls=None)
        ds = ds.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)  # this number can be tuned
        return ds

@dataset_registry(dataset_type="VOCRecord", framework="tensorflow, inteltensorflow, \
                  tensorflow_itex", dataset_format='')
class TensorflowVOCRecord(IterableDataset):
    """Configuration for PASCAL VOC 2012 database in tf record format

    Please arrange data in this way:
        root/val-00000-of-00004.tfrecord
        root/val-00001-of-00004.tfrecord
        ...
        root/val-00003-of-00004.tfrecord
    The file name needs to follow this pattern: 'val-*-of-*'

    Args: root (str): Root directory of dataset.
          transform (transform object, default=None):  transform to process input data.
          filter (Filter objects, default=None): filter out examples according
                                                 to specific conditions
    """
    def __new__(cls, root, transform=None, filter=None):

        from tensorflow.python.platform import gfile # pylint: disable=no-name-in-module
        glob_pattern = os.path.join(root, '%s-*' % 'val')
        file_names = gfile.Glob(glob_pattern)
        if not file_names:
            raise ValueError('Found no files in --root matching: {}'.format(glob_pattern))

        # pylint: disable=no-name-in-module
        from tensorflow.python.data.experimental import parallel_interleave
        ds = tf.data.TFRecordDataset.list_files(file_names, shuffle=False)
        ds = ds.apply(parallel_interleave(
                tf.data.TFRecordDataset, cycle_length=len(file_names)))

        if transform is not None:
            ds = ds.map(transform, num_parallel_calls=None)
        ds = ds.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)  # this number can be tuned
        return ds



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_util as tu
from ..graph_base import GraphRewriterBase
from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer
from neural_compressor.adaptor.tf_utils.graph_util import GraphRewriterHelper as Helper


class InsertPrintMinMaxNode(GraphRewriterBase):
    """InsertPrintMinMaxNode Pass for tensorflow sampling.
    """

    def __init__(self, model, pre_node_name, post_node_name, frame_name=None):
        super().__init__(model)
        self.pre_node_name = pre_node_name
        self.post_node_name = post_node_name
        self.signature = pre_node_name + post_node_name
        self.frame_name = frame_name

    def do_transformation(self):
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        insert_node_pairs = []
        top_node = graph_info[self.pre_node_name].node
        if top_node.op == 'ConcatV2':
            for i in range(top_node.attr['N'].i):
                insert_node_pairs.append([top_node.input[i], self.post_node_name])
        else:
            refresh_pre_node_name = graph_info[self.pre_node_name].node.input[0]
            # Check the Conv2D could be fused with previous Pad or not.
            # If so, we need to update the pre-node name correspondingly.
            refresh_pre_node = graph_info[Helper.node_name_from_input(refresh_pre_node_name)].node
            if refresh_pre_node.op == 'Pad' and( top_node.op == 'Conv2D' or top_node.op == 'Conv3D'):
                pad_const_node_name = refresh_pre_node.input[1]
                pad_const_node = graph_info[pad_const_node_name].node
                padding_tensor = tu.MakeNdarray(pad_const_node.attr["value"].tensor).flatten()
                if not any(padding_tensor) or \
                    (any(padding_tensor) and tf.version.VERSION in ( '1.15.0-up3', '2.8.0202151')):
                    refresh_pre_node_name = refresh_pre_node.input[0]

            insert_node_pairs.append([refresh_pre_node_name, self.post_node_name])

        output_names = []
        for node_pair_names in insert_node_pairs:
            for index, each_node_name in enumerate(node_pair_names):
                name_with_sig = each_node_name + self.signature
                node_name_prefix = name_with_sig.replace(":", "__port__").replace("^", "__hat__")
                reshape_dims_name = node_name_prefix + "_reshape_dims"
                reduction_dims_name = node_name_prefix + "_reduction_dims"

                reshape_dims_node = Helper.create_constant_node(
                    reshape_dims_name, -1, dtypes.int32, [1])

                reduction_dims_node = Helper.create_constant_node(
                    reduction_dims_name, 0, dtypes.int32, [1])

                if self.frame_name:         # pragma: no cover
                    reduction_dims_enter_node = Helper.create_node(
                        'Enter', reduction_dims_name+'_enter', [reduction_dims_name])
                    reshape_dims_enter_node = Helper.create_node(
                        'Enter', reshape_dims_name+'_enter', [reshape_dims_name])
                    Helper.set_attr_string(reduction_dims_enter_node,
                                           'frame_name', self.frame_name.encode())
                    Helper.set_attr_dtype(reduction_dims_enter_node, 'T', dtypes.int32)
                    # Helper.set_attr_bool(reduction_dims_enter_node, 'is_constant', True)
                    Helper.set_attr_int(reduction_dims_enter_node, 'parallel_iterations', 32)
                    Helper.set_attr_string(reshape_dims_enter_node,
                                           'frame_name', self.frame_name.encode())
                    Helper.set_attr_dtype(reshape_dims_enter_node, 'T', dtypes.int32)
                    # Helper.set_attr_bool(reshape_dims_enter_node, 'is_constant', False)
                    Helper.set_attr_int(reshape_dims_enter_node, 'parallel_iterations', 32)

                reshape_input_name = node_name_prefix + "_reshape_"

                if self.frame_name:
                    reshape_input_node = Helper.create_node("Reshape", reshape_input_name, [
                        each_node_name, reshape_dims_enter_node.name])
                else:
                    reshape_input_node = Helper.create_node("Reshape", reshape_input_name, [
                                                            each_node_name, reshape_dims_name])

                min_input_name = node_name_prefix + "_min"
                if self.frame_name:
                    min_input_node = Helper.create_node(
                        "Min", min_input_name,
                        [reshape_input_name, reduction_dims_enter_node.name])
                else:
                    min_input_node = Helper.create_node(
                        "Min", min_input_name, [reshape_input_name, reduction_dims_name])
                Helper.set_attr_dtype(min_input_node, "Tidx", dtypes.int32)
                Helper.set_attr_bool(min_input_node, "keep_dims", False)

                max_input_name = node_name_prefix + "_max"
                if self.frame_name:
                    max_input_node = Helper.create_node(
                        "Max", max_input_name,
                        [reshape_input_name, reduction_dims_enter_node.name])
                else:
                    max_input_node = Helper.create_node(
                        "Max", max_input_name, [reshape_input_name, reduction_dims_name])
                Helper.set_attr_dtype(max_input_node, "Tidx", dtypes.int32)
                Helper.set_attr_bool(max_input_node, "keep_dims", False)

                max_print_node = Helper.create_node(
                    "Print", node_name_prefix + "_print_max__{}".format(index),
                    [max_input_name + ':0', max_input_name+':0'])
                min_print_node = Helper.create_node(
                    "Print", node_name_prefix + "_print_min__{}".format(index),
                    [min_input_name+':0', min_input_name+':0'])

                if index == 0:
                    max_msg = ';{}_eightbit_max_{}__print__;__max:'.format(
                        self.pre_node_name, each_node_name)
                    min_msg = ';{}_eightbit_min_{}__print__;__min:'.format(
                        self.pre_node_name, each_node_name)
                    src_dt = graph_info[self.pre_node_name].node.attr["T"]
                else:
                    max_msg = ';{}_eightbit_requant_range__print__;__requant_max:'.format(
                        self.pre_node_name)
                    min_msg = ';{}_eightbit_requant_range__print__;__requant_min:'.format(
                        self.pre_node_name)
                    src_dt = graph_info[each_node_name].node.attr["T"]

                if src_dt.type==0:
                    #workround for swish_f32, attribute T is not in the op definition
                    src_dt=attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum)

                reshape_input_node.attr["T"].CopyFrom(src_dt)
                min_input_node.attr["T"].CopyFrom(src_dt)
                min_print_node.attr["T"].CopyFrom(src_dt)
                max_input_node.attr["T"].CopyFrom(src_dt)
                max_print_node.attr["T"].CopyFrom(src_dt)

                min_print_node.attr["message"].s = min_msg.encode()
                min_print_node.attr["first_n"].i = -1
                min_print_node.attr["summarize"].i = 1024

                max_print_node.attr["message"].s = max_msg.encode()
                max_print_node.attr["first_n"].i = -1
                max_print_node.attr["summarize"].i = 1024

                attr_u = [dtypes.as_dtype(src_dt.type).as_datatype_enum]
                min_print_node.attr["U"].list.CopyFrom(
                    attr_value_pb2.AttrValue.ListValue(type=attr_u))
                max_print_node.attr["U"].list.CopyFrom(
                    attr_value_pb2.AttrValue.ListValue(type=attr_u))
                if self.frame_name:
                    cur_graph.add_node(reshape_dims_node, None, [reshape_dims_enter_node.name])
                    cur_graph.add_node(reduction_dims_node, None, [reduction_dims_enter_node.name])
                    cur_graph.add_node(reshape_dims_enter_node, None, [reshape_input_name])
                    cur_graph.add_node(reduction_dims_enter_node, None,
                                       [max_input_name, min_input_name])
                else:
                    cur_graph.add_node(reshape_dims_node, None, [reshape_input_name])
                    cur_graph.add_node(reduction_dims_node, None, [max_input_name, min_input_name])
                cur_graph.add_node(max_input_node, None, [max_print_node.name])
                cur_graph.add_node(min_input_node, None, [min_print_node.name])
                cur_graph.add_node(min_print_node, min_input_name, [])
                cur_graph.add_node(max_print_node, max_input_name, [])

                cur_graph.add_node(reshape_input_node, each_node_name,
                                   [max_input_name, min_input_name])
                output_names.append(max_print_node.name)
                output_names.append(min_print_node.name)

        return cur_graph.dump_graph(), output_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neural_compressor.utils.utility import dump_elapsed_time

from ..graph_base import GraphRewriterBase

class StripUnusedNodesOptimizer(GraphRewriterBase):
    def __init__(self, model, input_node_names, output_node_names):
        super().__init__(model)
        self.input_node_names = input_node_names
        self.output_node_names = output_node_names

    @dump_elapsed_time("Pass StripUnusedNodesOptimizer")
    def do_transformation(self):
        from neural_compressor.adaptor.tf_utils.util import fix_ref_type_of_graph_def
        from neural_compressor.adaptor.tf_utils.util import strip_unused_nodes
        self.model = fix_ref_type_of_graph_def(self.model)
        return strip_unused_nodes(self.model,
                                  self.input_node_names,
                                  self.output_node_names)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import os

import torch
import transformers
from transformers import AutoConfig
from transformers.file_utils import cached_path, hf_bucket_url

from neural_compressor.utils import logger
from neural_compressor.utils.pytorch import load

WEIGHTS_NAME = "pytorch_model.bin"


class OptimizedModel:
    def __init__(self, *args, **kwargs):
        raise EnvironmentError(
            f"{self.__class__.__name__} is designed to be instantiated using the"
            f"`{self.__class__.__name__}.from_pretrained(model_name_or_path)` method."
        )

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        **kwargs
    ) -> torch.nn.Module:
        """
        Instantiate a quantized pytorch model from a given Intel Neural Compressor (INC) configuration file.
        Args:
            model_name_or_path (:obj:`str`):
                Repository name in the Hugging Face Hub or path to a local directory hosting the model.
            cache_dir (:obj:`str`, `optional`):
                Path to a directory in which a downloaded configuration should be cached if the standard cache should
                not be used.
            force_download (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether or not to force to (re-)download the configuration files and override the cached versions if
                they exist.
            resume_download (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether or not to delete incompletely received file. Attempts to resume the download if such a file
                exists.
            revision(:obj:`str`, `optional`):
                The specific model version to use. It can be a branch name, a tag name, or a commit id, since we use a
                git-based system for storing models and other artifacts on huggingface.co, so ``revision`` can be any
                identifier allowed by git.
        Returns:
            q_model: Quantized model.
        """
        config = kwargs.pop("config", None)
        cache_dir = kwargs.pop("cache_dir", None)
        force_download = kwargs.pop("force_download", False)
        resume_download = kwargs.pop("resume_download", False)
        use_auth_token = kwargs.pop("use_auth_token", None)
        revision = kwargs.pop("revision", None)

        if config is None:
            config = AutoConfig.from_pretrained(
                model_name_or_path,
                cache_dir=cache_dir,
                force_download=force_download,
                resume_download=resume_download,
                use_auth_token=use_auth_token,
                revision=revision,
                **kwargs,
                )

        model_class = eval(f'transformers.{config.architectures[0]}')
        if config.torch_dtype is not torch.int8:
            logger.info("the prunging/distillation optimized model is loading.")
            model = model_class.from_pretrained(model_name_or_path)
            return model
        else:
            logger.info("the quantization optimized model is loading.")
            keys_to_ignore_on_load_unexpected = copy.deepcopy(
                getattr(model_class, "_keys_to_ignore_on_load_unexpected", None)
            )
            keys_to_ignore_on_load_missing = \
                copy.deepcopy(getattr(model_class, "_keys_to_ignore_on_load_missing", None))

            # Avoid unnecessary warnings resulting from quantized model initialization
            quantized_keys_to_ignore_on_load = [r"zero_point", r"scale", 
                                                r"packed_params", r"constant", 
                                                r"module", r"best_configure"]
            if keys_to_ignore_on_load_unexpected is None:
                model_class._keys_to_ignore_on_load_unexpected = quantized_keys_to_ignore_on_load
            else:
                model_class._keys_to_ignore_on_load_unexpected.extend(
                    quantized_keys_to_ignore_on_load
                )
            missing_keys_to_ignore_on_load = [r"weight", r"bias"]
            if keys_to_ignore_on_load_missing is None:
                model_class._keys_to_ignore_on_load_missing = missing_keys_to_ignore_on_load
            else:
                model_class._keys_to_ignore_on_load_missing.extend(missing_keys_to_ignore_on_load)

            model = model_class.from_pretrained(
                model_name_or_path,
                cache_dir=cache_dir,
                force_download=force_download,
                resume_download=resume_download,
                use_auth_token=use_auth_token,
                revision=revision,
                **kwargs,
                )

            model_class._keys_to_ignore_on_load_unexpected = keys_to_ignore_on_load_unexpected
            model_class._keys_to_ignore_on_load_missing = keys_to_ignore_on_load_missing

            if not os.path.isdir(model_name_or_path) and not os.path.isfile(model_name_or_path):
                weights_file = hf_bucket_url(model_name_or_path,
                                            filename=WEIGHTS_NAME,
                                            revision=revision)
                try:
                    # Load from URL or cache if already cached
                    resolved_weights_file = cached_path(
                        weights_file,
                        cache_dir=cache_dir,
                        force_download=force_download,
                        resume_download=resume_download,
                        use_auth_token=use_auth_token,
                    )
                except EnvironmentError as err:
                    logger.error(err)
                    msg = (
                        f"Can't load weights for '{model_name_or_path}'. Make sure that:\n\n"
                        f"- '{model_name_or_path}' is a correct model identifier "
                        f"listed on 'https://huggingface.co/models'\n  (make sure "
                        f"'{model_name_or_path}' is not a path to a local directory with "
                        f"something else, in that case)\n\n- or '{model_name_or_path}' is "
                        f"the correct path to a directory containing a file "
                        f"named one of {WEIGHTS_NAME}\n\n"
                    )

                    if revision is not None:
                        msg += (f"- or '{revision}' is a valid git identifier "
                                f"(branch name, a tag name, or a commit id) that "
                                f"exists for this model name as listed on its model "
                                f"page on 'https://huggingface.co/models'\n\n"
                            )

                    raise EnvironmentError(msg)
                q_model = load(resolved_weights_file, model)
            else:
                weights_file = os.path.join(os.path.abspath(
                    os.path.expanduser(model_name_or_path)), WEIGHTS_NAME)
                q_model = load(weights_file, model)

            return q_model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import autograd.numpy as anp
import numpy as np
import pymoo
from pymoo.algorithms.moo.age import AGEMOEA
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.core.problem import Problem
from pymoo.factory import get_crossover, get_mutation, get_sampling
from pymoo.optimize import minimize

from neural_compressor.experimental.nas.dynast.dynas_utils import EvaluationInterface
from neural_compressor.utils import logger


class SearchAlgoManager:
    """
    Manages the search parameters for the DyNAS-T single/multi-objective search.

    Args:
        algorithm : string
            Define a multi-objective search algorithm.
        seed : int
            Seed value for pymoo search.
        verbose : Boolean
            Verbosity option
        engine : string
            Support different engine types (e.g. pymoo, optuna, etc.)
    """

    def __init__(
        self,
        algorithm: str = 'nsga2',
        seed: int = 0,
        verbose: bool = False,
        engine: str = 'pymoo',
    ) -> None:
        self.algorithm = algorithm
        self.seed = seed
        self.verbose = verbose
        self.engine = engine
        if self.algorithm == 'nsga2':
            self.configure_nsga2()
            self.engine = 'pymoo'
        elif self.algorithm == 'age':
            self.configure_age()
            self.engine = 'pymoo'
        else: # pragma: no cover
            logger.error(
                '[DyNAS-T] algorithm "{}" not implemented.'.format(self.algorithm)
            )
            raise NotImplementedError

    def configure_nsga2(
        self,
        population: int = 50,
        num_evals: int = 1000,
        warm_pop: np.ndarray = None,
        crossover_prob: float = 0.9,
        crossover_eta: float = 15.0,
        mutation_prob: float = 0.02,
        mutation_eta: float = 20.0,
    ) -> None:
        self.n_gens = num_evals / population

        if type(warm_pop) == 'numpy.ndarray':
            logger.info('[DyNAS-T] Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = get_sampling("int_lhs")

        self.algorithm_def = NSGA2(
            pop_size=population,
            sampling=sample_strategy,
            crossover=get_crossover("int_sbx", prob=crossover_prob, eta=crossover_eta),
            mutation=get_mutation("int_pm", prob=mutation_prob, eta=mutation_eta),
            eliminate_duplicates=True,
        )

    def configure_age(
        self,
        population: int = 50,
        num_evals: int = 1000,
        warm_pop: np.ndarray = None,
        crossover_prob: float = 0.9,
        crossover_eta: float = 15.0,
        mutation_prob: float = 0.02,
        mutation_eta: float = 20.0,
    ) -> None:
        self.engine = 'pymoo'
        self.n_gens = num_evals / population

        if type(warm_pop) == 'numpy.ndarray':
            logger.info('[DyNAS-T] Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = get_sampling("int_lhs")

        self.algorithm_def = AGEMOEA(
            pop_size=population,
            sampling=sample_strategy,
            crossover=get_crossover("int_sbx", prob=crossover_prob, eta=crossover_eta),
            mutation=get_mutation("int_pm", prob=mutation_prob, eta=mutation_eta),
            eliminate_duplicates=True,
        )

    def run_search(
        self,
        problem: Problem,
        save_history=False,
    ) -> pymoo.core.result.Result:
        """Starts the search process for the algorithm and problem class that have
        been previously defined.
        """

        logger.info('[DyNAS-T] Running Search')
        start_time = time.time()

        # Note: Known issue with the pymoo v0.5.0 save_history=True option
        if self.engine == 'pymoo':
            result = minimize(
                problem,
                self.algorithm_def,
                ('n_gen', int(self.n_gens)),
                seed=self.seed,
                save_history=save_history,
                verbose=self.verbose,
            )
        else: # pragma: no cover
            logger.error('[DyNAS-T] Invalid algorithm engine configuration!')
            raise NotImplementedError

        logger.info(
            '[DyNAS-T] Genetic algorithm search took {:.3f} seconds.'.format(
                time.time() - start_time
            )
        )

        return result


class ProblemMultiObjective(Problem):
    """
    Interface between the user-defined evaluation interface and the SearchAlgoManager.

    Args:
        evaluation_interface : Class
            Class that handles the objective measurement call from the supernet.
        param_count : int
            Number variables in the search space (e.g., OFA MobileNetV3 has 45)
        param_upperbound : array
            The upper int array that defines how many options each design variable has.
    """

    def __init__(
        self,
        evaluation_interface: EvaluationInterface,
        param_count: int,
        param_upperbound: list,
    ):
        super().__init__(
            n_var=param_count,
            n_obj=2,
            n_constr=0,
            xl=0,
            xu=param_upperbound,
            type_var=np.int,
        )

        self.evaluation_interface = evaluation_interface

    def _evaluate(
        self,
        x: list,
        out,
        *args,
        **kwargs,
    ) -> None:

        # Store results for a given generation for PyMoo
        objective_x_arr, objective_y_arr = list(), list()

        # Measure new individuals
        for i in range(len(x)):

            _, objective_x, objective_y = self.evaluation_interface.eval_subnet(x[i])

            objective_x_arr.append(objective_x)
            objective_y_arr.append(objective_y)

        print('.', end='', flush=True)

        # Update PyMoo with evaluation data
        out["F"] = anp.column_stack([objective_x_arr, objective_y_arr])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


NASMethods = {}


def nas_registry(nas_method):
    """The class decorator used to register all NAS subclasses.
    Args:
        nas_method (str): The string of supported NAS Method.

    Returns:
        cls: The class of register.
    """
    assert isinstance(nas_method, str), "Expect nas_method to be a string."
    def decorator(cls):
        NASMethods[nas_method.lower()] = cls
        return cls
    return decorator


def create_search_space_pool(search_space, idx=0):
    search_space_keys = sorted(search_space.keys())
    if idx == len(search_space_keys):
        return [[]]
    key = search_space_keys[idx]
    search_space_pool = []
    for v in search_space[key]:
        sub_search_space_pool = create_search_space_pool(search_space, idx+1)
        search_space_pool += [[v] + item for item in sub_search_space_pool]
    return search_space_pool


def find_pareto_front(metrics):
    """
    Find the pareto front points, assuming all metrics are "higher is better".

    Args:
        metrics: An (n_points, n_metrics) array
    Return:
        An array of indices of pareto front points.
        It is a (n_pareto_points, ) integer array of indices.
    """
    metrics = np.array(metrics)
    pareto_front_point_indices = np.arange(metrics.shape[0])
    next_point_idx = 0
    while next_point_idx < len(metrics):
        nondominated_points = np.any(metrics > metrics[next_point_idx], axis=1)
        nondominated_points[next_point_idx] = True
        # Remove points being dominated by current point
        pareto_front_point_indices = pareto_front_point_indices[nondominated_points]
        metrics = metrics[nondominated_points]
        next_point_idx = np.sum(nondominated_points[:next_point_idx+1])
    return pareto_front_point_indices


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class CastOperatorBase:
    def __init__(self, onnx_quantizer, onnx_node):
        self.quantizer = onnx_quantizer
        self.node = onnx_node
        self.dtype = self.quantizer.config[self.node.name]

    def cast(self):
        self.quantizer.dtype_cast(self.node, self.dtype)



#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
from torch.fx import symbolic_trace

class BF16ModuleWrapper(nn.Module):
    def __init__(self, module):
        super(BF16ModuleWrapper, self).__init__()
        self.add_module('module', module)
        self.train(module.training)

    def forward(self, X):
        X = X.to(torch.bfloat16)
        self.module.bfloat16()
        X = self.module(X)
        return X.float()

def Convert(model, tune_cfg):
        bf16_ops_list = tune_cfg['bf16_ops_list']
        fx_sub_module_list = tune_cfg['fx_sub_module_list'] \
                             if 'fx_sub_module_list' in tune_cfg.keys() else []
        mixed_precision_model = bf16_wrapper_model(model, bf16_ops_list)
        if len(fx_sub_module_list) > 0:
            mixed_precision_model = bf16_symbolic_trace(mixed_precision_model, fx_sub_module_list)
        return mixed_precision_model

def bf16_wrapper_model(model, bf16_ops_list, prefix=''):
    for name, child in model.named_children():
        op_name = prefix + '.' + name if prefix != '' else name
        for bf16_op_name in bf16_ops_list:
            if op_name == bf16_op_name[0]:
                child = BF16ModuleWrapper(child)
        else:
            bf16_wrapper_model(child, bf16_ops_list, op_name)
            setattr(model, name, child)
    return model


def bf16_symbolic_trace(model, fx_sub_module_list, prefix=''):
    for name, child in model.named_children():
        op_name = prefix + '.' + name if prefix != '' else name
        for fx_sub_module_name in fx_sub_module_list:
            if op_name == fx_sub_module_name:
                child = symbolic_trace(child)
        else:
            bf16_symbolic_trace(child, fx_sub_module_list, op_name)
            setattr(model, name, child)
    return model


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import random
import tempfile
import sys
import datetime
import numpy as np
import yaml
from neural_compressor.adaptor import FRAMEWORKS
from ..conf.config import Conf
from ..conf.dotdict import deep_get, deep_set, DotDict
from ..strategy import STRATEGIES
from ..utils import logger
from ..utils.create_obj_from_config import create_dataloader, create_eval_func
from ..utils.utility import CpuInfo, set_backend
from .common import Model as NCModel
from ..model import BaseModel

class ModelConversion():
    """ModelConversion class is used to convert one model format to another.

       Currently Neural Compressor only supports Quantization-aware training TensorFlow model to Default
       quantized model.

       The typical usage is:
         from neural_compressor.experimental import ModelConversion, common
         conversion = ModelConversion()
         conversion.source = 'QAT'
         conversion.destination = 'default'
         conversion.model = '/path/to/saved_model'
         q_model = conversion()

    Args:
        conf_fname_or_obj (string or obj): Optional. The path to the YAML configuration file or
            Conf class containing model conversion and evaluation setting if not specifed by code.

    """

    def __init__(self, conf_fname_or_obj=None):
        self.conf_name = conf_fname_or_obj
        self._model = None
        self.framework = 'tensorflow'

        self._eval_dataloader = None
        self._eval_func = None
        self.adaptor = None
        self._metric = None

        self._source = None
        self._destination = None

        if conf_fname_or_obj is not None:
            if isinstance(conf_fname_or_obj, str):
                self.conf = Conf(conf_fname_or_obj)
            elif isinstance(conf_fname_or_obj, Conf):
                self.conf = conf_fname_or_obj
            else:  # pragma: no cover
                assert False, \
                    "Please pass a YAML configuration file path or \
                    Conf class to model_conversion"
        else:
            self.conf = None

        set_backend(self.framework)

    def __call__(self):
        """The main entry point of model conversion process.

           NOTE: This interface works now only on Intel Optimized TensorFlow to
           convert QAT model generated by tensorflow_model_optimization to default
           quantized model which is able to run at Intel Xeon platforms.
           https://github.com/tensorflow/model-optimization

        Returns:
            converted quantized model

        """
        assert self._model, '"model" property need to be set before __call_() gets invoked'

        framework_specific_info = {}
        cfg = self.conf.usr_cfg
        framework_specific_info.update(
            {'name': cfg.model.name,
             'device': cfg.device,
             'fake_quant': True,
             'inputs': cfg.model.inputs,
             'outputs': cfg.model.outputs,
             'workspace_path': cfg.tuning.workspace.path})

        self.adaptor = FRAMEWORKS[self.framework](framework_specific_info)
        q_model = self.adaptor.convert(self._model, self._source, self._destination)

        # when eval_func is None but metric or _eval_dataloader is set by yaml or code,
        # it means Neural Compressor will create the eval_func from these info.
        metric_cfg = [self._metric] if self._metric else \
            deep_get(cfg, 'evaluation.accuracy.metric')
        postprocess_cfg = deep_get(cfg, 'evaluation.accuracy.postprocess')
        if self._eval_func is None and metric_cfg:
            eval_dataloader_cfg = deep_get(cfg, 'evaluation.accuracy.dataloader')
            if self._eval_dataloader is None and eval_dataloader_cfg:
                self._eval_dataloader = create_dataloader(self.framework, eval_dataloader_cfg)
            assert self._eval_dataloader, 'either "eval_dataloader" property or evaluation' \
                    '.accuracy.dataloader field in yaml should be set when metric is set'

            self._eval_func = create_eval_func(self.framework, \
                                            self.eval_dataloader, \
                                            self.adaptor, \
                                            metric_cfg, \
                                            postprocess_cfg, \
                                            fp32_baseline = True)
        if self._eval_func:
            baseline_score = self._eval_func(self._model)
            qmodel_score = self._eval_func(q_model)
            logger.info("The score of Quantization-Aware Training model is {}.".
                        format(str(baseline_score)))
            logger.info("Converted model score is {}.".format(str(qmodel_score)))

        return q_model

    fit = __call__

    def _gen_yaml(self):
        random_name = '{}'.format(datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
        default_yaml_template = {'model': {'framework': self.framework, 'name': random_name},
                                 'device': 'cpu',
                                 'model_conversion': {'source': 'QAT', 'destination': 'default'}}

        temp_yaml_path = tempfile.mkstemp(suffix='.yaml')[1]
        with open(temp_yaml_path, 'w', encoding='utf-8') as f:
            yaml.dump(default_yaml_template, f)
        self.conf = Conf(temp_yaml_path)

    def dataset(self, dataset_type, *args, **kwargs):
        from .data import DATASETS
        return DATASETS(self.framework)[dataset_type](*args, **kwargs)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, _source):
        assert _source.lower() == 'qat', 'Model conversion now only supports TensorFlow ' \
                                         'QAT model to default quantized model'
        self._source = _source.lower()

    @property
    def destination(self):
        return self._destination

    @destination.setter
    def destination(self, _destination):
        assert _destination.lower() == 'default', 'Model conversion now only supports ' \
                                       'TensorFlow QAT model to default quantized model'
        self._destination = _destination.lower()

    @property
    def eval_dataloader(self):
        return self._eval_dataloader

    @eval_dataloader.setter
    def eval_dataloader(self, dataloader):
        """Set Data loader for evaluation, It is iterable and the batched data
           should consists of a tuple like (input, label), when eval_dataloader is set,
           user should configure postprocess(optional) and metric in yaml file or set
           postprocess and metric cls. Notice evaluation dataloader will be used to
           generate data for model inference, make sure the input data can be feed to model.

           Args:
               dataloader(generator): user are supported to set a user defined dataloader
                                      which meet the requirements that can yield tuple of
                                      (input, label)/(input, _) batched data.
                                      Another good practice is to use neural_compressor.common.DataLoader
                                      to initialize a neural_compressor dataloader object.
                                      Notice neural_compressor.common.DataLoader is just a wrapper of the
                                      information needed to build a dataloader, it can't yield
                                      batched data and only in this setter method
                                      a 'real' eval_dataloader will be created,
                                      the reason is we have to know the framework info
                                      and only after the Quantization object created then
                                      framework infomation can be known. Future we will support
                                      creating iterable dataloader from neural_compressor.common.DataLoader

        """
        from .common import _generate_common_dataloader
        self._eval_dataloader = _generate_common_dataloader(
            dataloader, self.framework)

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, user_model):
        """Set the user model and dispatch to framework specific internal model object

        Args:
           user_model: user are supported to set model from original framework model format
                       (eg, tensorflow frozen_pb or path to a saved model), but not recommended.
                       Best practice is to set from a initialized neural_compressor.common.Model.
                       If tensorflow model is used, model's inputs/outputs will be auto inferred,
                       but sometimes auto inferred inputs/outputs will not meet your requests,
                       set them manually in config yaml file. Another corner case is slim model
                       of tensorflow, be careful of the name of model configured in yaml file,
                       make sure the name is in supported slim model list.

        """
        if not isinstance(user_model, BaseModel):
            logger.warning("Force convert framework model to neural_compressor model.")
            self._model = NCModel(user_model)
        else:
            self._model = user_model

        assert self.framework == 'tensorflow', \
            'Model conversion only supports Tensorflow at current stage.'

        if not self.conf:
            self._gen_yaml()

        cfg = self.conf.usr_cfg
        if self.framework == 'tensorflow':
            self._model.name = cfg.model.name
            self._model.workspace_path = cfg.tuning.workspace.path

    @property
    def metric(self):
        assert False, 'Should not try to get the value of `metric` attribute.'
        return None

    @metric.setter
    def metric(self, user_metric):
        """Set metric class and neural_compressor will initialize this class when evaluation
           neural_compressor have many built-in metrics, but user can set specific metric through
           this api. The metric class should take the outputs of the model or
           postprocess(if have) as inputs, neural_compressor built-in metric always take
           (predictions, labels) as inputs for update,
           and user_metric.metric_cls should be sub_class of neural_compressor.metric.BaseMetric
           or user defined metric object

        Args:
            user_metric(neural_compressor.common.Metric): user_metric should be object initialized from
                                             neural_compressor.common.Metric, in this method the
                                             user_metric.metric_cls will be registered to
                                             specific frameworks and initialized.

        """
        if deep_get(self.conf.usr_cfg, "evaluation.accuracy.metric"):
            logger.warning("Override the value of `metric` field defined in yaml file" \
                           " as user defines the value of `metric` attribute by code.")
 
        from .common import Metric as NCMetric
        if isinstance(user_metric, NCMetric):
            metric_cfg = {user_metric.name : {**user_metric.kwargs}}
            deep_set(self.conf.usr_cfg, "evaluation.accuracy.metric", metric_cfg)
            self.conf.usr_cfg = DotDict(self.conf.usr_cfg)
            from .metric import METRICS
            metrics = METRICS(self.framework)
            metrics.register(user_metric.name, user_metric.metric_cls)
        else:
            for i in ['reset', 'update', 'result']:
                assert hasattr(user_metric, i), 'Please realise {} function' \
                                                'in user defined metric'.format(i)
            self._metric = user_metric

    @property
    def postprocess(self, user_postprocess):
        assert False, 'Should not try to get the value of `postprocess` attribute.'
        return None

    @postprocess.setter
    def postprocess(self, user_postprocess):
        """Set postprocess class and neural_compressor will initialize this class when evaluation.
           The postprocess class should take the outputs of the model as inputs, and
           output (predictions, labels) as inputs for metric update.
           user_postprocess.postprocess_cls should be sub_class of neural_compressor.data.BaseTransform.

        Args:
            user_postprocess(neural_compressor.common.Postprocess):
                user_postprocess should be object initialized from neural_compressor.common.Postprocess,
                in this method the user_postprocess.postprocess_cls will be
                registered to specific frameworks and initialized.

        """
        from .common import Postprocess as NCPostprocess
        assert isinstance(user_postprocess, NCPostprocess), \
            'please initialize a neural_compressor.common.Postprocess and set....'
        postprocess_cfg = {user_postprocess.name : {**user_postprocess.kwargs}}
        if deep_get(self.conf.usr_cfg, "evaluation.accuracy.postprocess"):
            logger.warning("Override the value of `postprocess` field defined in yaml file" \
                           " as user defines the value of `postprocess` attribute by code.")
        deep_set(self.conf.usr_cfg, "evaluation.accuracy.postprocess.transform", postprocess_cfg)
        from .data import TRANSFORMS
        postprocesses = TRANSFORMS(self.framework, 'postprocess')
        postprocesses.register(user_postprocess.name, user_postprocess.postprocess_cls)

    @property
    def eval_func(self):
        assert False, 'Should not try to get the value of `eval_func` attribute.'
        return None

    @eval_func.setter
    def eval_func(self, user_eval_func):
        """ The evaluation function provided by user.

        Args:
            user_eval_func: This function takes model as parameter,
                            and evaluation dataset and metrics should be
                            encapsulated in this function implementation
                            and outputs a higher-is-better accuracy scalar
                            value.

                            The pseudo code should be something like:

                            def eval_func(model):
                                 input, label = dataloader()
                                 output = model(input)
                                 accuracy = metric(output, label)
                                 return accuracy
        """
        self._eval_func = user_eval_func

    def __repr__(self):
        return 'ModelConversion'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ..conf.config import Conf
from ..utils import logger
from ..utils.utility import set_backend
from ..utils.create_obj_from_config import create_dataloader, create_train_func, create_eval_func
from ..model import BaseModel
from .common import Model
from ..adaptor import FRAMEWORKS
from ..model.model import get_model_fwk_name
from warnings import warn


class Component(object):
    """This is base class of Neural Compressor Component
    """
    def __init__(self, conf_fname_or_obj=None, combination=None):
        self.conf = None
        self.cfg = None
        self.combination = combination
        self.framework = None
        self._model = None
        self._train_func = None
        self._train_dataloader = None
        self._eval_func = None
        self._eval_dataloader = None
        self._train_distributed = False
        self._evaluation_distributed = False
        self.adaptor = None
        self._metric = None
        self.hooks = {
            'on_train_begin': self.on_train_begin,
            'on_train_end': self.on_train_end,
            'on_epoch_begin': self.on_epoch_begin,
            'on_epoch_end': self.on_epoch_end,
            'on_step_begin': self.on_step_begin,
            'on_step_end': self.on_step_end,
            'on_after_compute_loss': self.on_after_compute_loss,
            'on_before_optimizer_step': self.on_before_optimizer_step
        }
        self.hooks_dict = {
            'on_train_begin': [],
            'on_train_end': [],
            'on_epoch_begin': [],
            'on_epoch_end': [],
            'on_step_begin': [],
            'on_step_end': [],
            'on_after_compute_loss': [],
            'on_before_optimizer_step': []
        }
        if conf_fname_or_obj is not None:  # pragma: no cover
            if isinstance(conf_fname_or_obj, str):
                self.conf = Conf(conf_fname_or_obj)
            elif isinstance(conf_fname_or_obj, Conf):
                self.conf = conf_fname_or_obj
            else:
                assert False, \
                    "Please pass a YAML configuration file path and name or \
                    Conf class to Component"
            self._init_with_conf()


    def _init_with_conf(self):
        self.cfg = self.conf.usr_cfg
        if self.cfg.model.framework != 'NA':
            self.framework = self.cfg.model.framework.lower()
            set_backend(self.framework)

    def pre_process(self):
        """ Initialize the dataloader and train/eval functions from yaml config.
            Component base class provides default function to initialize dataloaders and functions
            from user config. And for derived classes(Pruning, Quantization, etc.), an override
            function is required.
        """
        # create adaptor
        framework_specific_info = {'device': self.cfg.device,
                                   'random_seed': self.cfg.tuning.random_seed,
                                   'workspace_path': self.cfg.tuning.workspace.path,
                                   'q_dataloader': None}
        if self.cfg.quantization.approach is not None:
            framework_specific_info['approach'] = self.cfg.quantization.approach

        if 'tensorflow' in self.framework:
            framework_specific_info.update(
                {"inputs": self.cfg.model.inputs, "outputs": self.cfg.model.outputs})

        self.adaptor = FRAMEWORKS[self.framework](framework_specific_info)
        self.adaptor.model = self.model

        # create dataloaders
        if self._train_dataloader is None and self._train_func is None:
            train_dataloader_cfg = self.cfg.train.dataloader
            assert train_dataloader_cfg is not None, \
                   'No training dataloader setting in current component. Please check ' \
                   'dataloader field of train field in yaml file. Or manually pass ' \
                   'dataloader to component.'

            self._train_dataloader = create_dataloader(self.framework, train_dataloader_cfg)
        if self._eval_dataloader is None and self._eval_func is None:
            if self._eval_dataloader is None:
                eval_dataloader_cfg = self.cfg.evaluation.accuracy.dataloader
                assert eval_dataloader_cfg is not None, \
                   'No evaluation dataloader setting in current component. Please check ' \
                   'dataloader field of evaluation field in yaml file. Or manually pass ' \
                   'dataloader to component.'
                self._eval_dataloader = create_dataloader(self.framework, eval_dataloader_cfg)

        # create functions
        if self._train_func is None:
            self._train_func = create_train_func(self.framework,
                                                    self._train_dataloader,
                                                    self.adaptor,
                                                    self.cfg.train,
                                                    hooks=self.hooks)
        if self._eval_func is None:
            metric = [self._metric] if self._metric else self.cfg.evaluation.accuracy.metric
            self._eval_func = create_eval_func(self.framework,
                                               self._eval_dataloader,
                                               self.adaptor,
                                               metric,
                                               self.cfg.evaluation.accuracy.postprocess,
                                               fp32_baseline = False)
        # register Quantization Aware Training hooks
        if self.combination is not None and 'Quantization' in self.combination:
            self.register_hook('on_train_begin', self.adaptor._pre_hook_for_qat)
            self.register_hook('on_train_end', self.adaptor._post_hook_for_qat)
        # strategy will be considered in future
        if getattr(self.train_dataloader, 'distributed', False):
            self.register_hook('on_train_begin', self.adaptor._pre_hook_for_hvd)

    def execute(self):
        """ Initialize the dataloader and train/eval functions from yaml config.
            Component base class provides default function to initialize dataloaders and functions
            from user config. And for derived classes(Pruning, Quantization, etc.), an override
            function is required.
        """
        # TODO: consider strategy sync during combination
        if self._train_func is not None:
            modified_model = self._train_func(self._model \
                    if getattr(self._train_func, 'builtin', None) else self._model.model)
            # for the cases that model is changed not inplaced during training, for example,
            # oneshot with torch_fx QAT interfaces. Needs to reset model afterwards.
            if modified_model is not None:
                self._model.model = modified_model
        if self._eval_func is not None:
            score = self._eval_func(self._model \
                    if getattr(self._eval_func, 'builtin', None) else self._model.model)
            logger.info("Evaluated model score is {}.".format(str(score)))
        return self._model

    def post_process(self):
        pass

    def on_train_begin(self):
        """ called before the beginning of epochs"""
        for on_train_begin_hook in self.hooks_dict['on_train_begin']:
            on_train_begin_hook()

    def on_train_end(self):
        """ called after the end of epochs"""
        for on_train_end_hook in self.hooks_dict['on_train_end']:
            on_train_end_hook()

    def pre_epoch_begin(self):
        """ called before the beginning of epochs"""
        warn('This method is deprecated. please use `on_train_begin` instead.',
             DeprecationWarning, stacklevel=2)
        for on_train_begin_hook in self.hooks_dict['on_train_begin']:
            on_train_begin_hook()

    def post_epoch_end(self):
        """ called after the end of epochs"""
        warn('This method is deprecated. please use `on_train_end` instead.',
             DeprecationWarning, stacklevel=2)
        for on_train_end_hook in self.hooks_dict['on_train_end']:
            on_train_end_hook()

    def on_epoch_begin(self, epoch):
        """ called on the beginning of epochs"""
        for on_epoch_begin_hook in self.hooks_dict['on_epoch_begin']:
            on_epoch_begin_hook(epoch)

    def on_step_begin(self, batch_id):
        """ called on the beginning of batches"""
        res_list = []
        for on_step_begin_hook in self.hooks_dict['on_step_begin']:
            res_list.append(on_step_begin_hook(batch_id))
        return res_list

    def on_batch_begin(self, batch_id):
        warn('This method is deprecated. please use `on_step_begin` instead.',
             DeprecationWarning, stacklevel=2)
        return self.on_step_begin(batch_id)

    def on_after_compute_loss(self, input, student_output, student_loss, teacher_output=None):
        """ called on the end of loss computation"""
        loss = student_loss
        for on_after_compute_loss_hook in self.hooks_dict['on_after_compute_loss']:
            loss = on_after_compute_loss_hook(input, student_output, loss, teacher_output)
        return loss

    def on_before_optimizer_step(self):
        """ called on the end of backward"""
        for on_before_optimizer_step_hook in self.hooks_dict['on_before_optimizer_step']:
            on_before_optimizer_step_hook()

    def on_post_grad(self):
        warn('This method is deprecated. please use `on_before_optimizer_step` instead.',
             DeprecationWarning, stacklevel=2)
        return self.on_before_optimizer_step()

    def on_step_end(self):
        """ called on the end of batches"""
        res_list = []
        for on_step_end_hook in self.hooks_dict['on_step_end']:
            res_list.append(on_step_end_hook())
        return res_list

    def on_batch_end(self):
        warn('This method is deprecated. please use `on_step_end` instead.',
             DeprecationWarning, stacklevel=2)
        return self.on_step_end()

    def on_epoch_end(self):
        """ called on the end of epochs"""
        res_list = []

        for on_epoch_end_hook in self.hooks_dict['on_epoch_end']:
            res_list.append(on_epoch_end_hook())

        return res_list

    def register_hook(self, scope, hook, input_args=None, input_kwargs=None):
        """ register hook for component. input_args and input_kwargs are reserved for user
            registered hooks."""
        if hook not in self.hooks_dict[scope]:
            self.hooks_dict[scope].append(hook)

    def __call__(self):
        self.pre_process()
        results = self.execute()
        self.post_process()
        return results

    def __repr__(self):
        if self.combination:
            return 'Combination of ' + ','.join(self.combination)
        else:
            return 'Base Component'

    @property
    def train_func(self):
        """ not support get train_func """
        assert False, 'Should not try to get the value of `train_func` attribute.'
        return None

    @train_func.setter
    def train_func(self, user_train_func):
        """Training function.

        Args:
            user_training_func: This function takes "model" as input parameter
                         and executes entire training process with self
                         contained training hyper-parameters. If training_func set,
                         an evaluation process must be triggered and user should
                         set eval_dataloader with metric configured or directly eval_func
                         to make evaluation of the model executed. training_func will return
                         a trained model.
        """
        self._train_func = user_train_func

    @property
    def eval_func(self):
        """ not support get eval_func """
        assert False, 'Should not try to get the value of `eval_func` attribute.'
        return None

    @eval_func.setter
    def eval_func(self, user_eval_func):
        """Eval function for component.

        Args:
            user_eval_func: This function takes "model" as input parameter
                         and executes entire evaluation process with self
                         contained metrics. If eval_func set,
                         an evaluation process must be triggered 
                         to make evaluation of the model executed.
        """
        self._eval_func = user_eval_func

    @property
    def train_dataloader(self):
        """ Getter to train dataloader """
        return self._train_dataloader

    @train_dataloader.setter
    def train_dataloader(self, dataloader):
        """Set Data loader for training for Component.
           It is iterable and the batched data should consists of a tuple like
           (input, label) if the training dataset containing label, or yield (input, _)
           for label-free train dataset, the input in the batched data will be used for
           model inference, so it should satisfy the input format of specific model.
           In train process, label of data loader will not be used and
           neither the postprocess and metric. User only need to set
           train_dataloader when train_dataloader can not be configured from yaml file.

           Args:
               dataloader(generator): user are supported to set a user defined dataloader
                                      which meet the requirements that can yield tuple of
                                      (input, label)/(input, _) batched data. Another good
                                      practice is to use neural_compressor.experimental.common.DataLoader
                                      to initialize a neural_compressor dataloader object. Notice
                                      neural_compressor.experimental.common.DataLoader is just a wrapper of the
                                      information needed to build a dataloader, it can't yield
                                      batched data and only in this setter method
                                      a 'real' train_dataloader will be created,
                                      the reason is we have to know the framework info
                                      and only after the Component object created then
                                      framework information can be known.
                                      Future we will support creating iterable dataloader
                                      from neural_compressor.experimental.common.DataLoader
        """
        from .common import _generate_common_dataloader
        self._train_dataloader = _generate_common_dataloader(
            dataloader, self.framework, self._train_distributed)

    @property
    def eval_dataloader(self):
        """ Getter to eval dataloader """
        return self._eval_dataloader

    @eval_dataloader.setter
    def eval_dataloader(self, dataloader):
        """Set Data loader for evaluation of component.
           It is iterable and the batched data should consists of yield (input, _).
           the input in the batched data will be used for model inference, so it
           should satisfy the input format of specific model.
           User only need to set eval_dataloader when eval_dataloader can not be
           configured from yaml file.

           Args:
               dataloader(generator): user are supported to set a user defined dataloader
                                      which meet the requirements that can yield tuple of
                                      (input, label)/(input, _) batched data. Another good
                                      practice is to use neural_compressor.experimental.common.DataLoader
                                      to initialize a neural_compressor dataloader object. Notice
                                      neural_compressor.experimental.common.DataLoader is just a wrapper of the
                                      information needed to build a dataloader, it can't yield
                                      batched data and only in this setter method
                                      a 'real' train_dataloader will be created,
                                      the reason is we have to know the framework info
                                      and only after the Component object created then
                                      framework information can be known.
                                      Future we will support creating iterable dataloader
                                      from neural_compressor.experimental.common.DataLoader
        """
        from .common import _generate_common_dataloader
        self._eval_dataloader = _generate_common_dataloader(
            dataloader, self.framework, self._evaluation_distributed)

    @property
    def model(self):
        """ Getter of model in neural_compressor.model  """
        return self._model

    @model.setter
    def model(self, user_model):
        """Set the user model and dispatch to framework specific internal model object

        Args:
           user_model: user are supported to set model from original framework model format
                       (eg, tensorflow frozen_pb or path to a saved model),
                       but not recommended. Best practice is to set from a initialized
                       neural_compressor.experimental.common.Model.
                       If tensorflow model is used, model's inputs/outputs will be
                       auto inferenced, but sometimes auto inferenced
                       inputs/outputs will not meet your requests,
                       set them manually in config yaml file.
                       Another corner case is slim model of tensorflow,
                       be careful of the name of model configured in yaml file,
                       make sure the name is in supported slim model list.

        """
        if not isinstance(user_model, BaseModel):
            logger.warning("Force convert framework model to neural_compressor model.")
            self._model = Model(user_model)
        else:
            self._model = user_model

        if self.cfg.model.framework == 'NA':
            self.framework = get_model_fwk_name(user_model)
            if self.framework == 'onnxruntime':
                self.framework = 'onnxrt_qoperator'
            self.cfg.model.framework = self.framework
            set_backend(self.framework)

        if 'tensorflow' in self.framework:
            self._model.name = self.cfg.model.name
            self._model.output_tensor_names = self.cfg.model.outputs
            self._model.input_tensor_names = self.cfg.model.inputs
            self._model.workspace_path = self.cfg.tuning.workspace.path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
import numpy
from .base_operator import QuantOperatorBase
from .qdq_base_operator import QDQOperatorBase
from onnxruntime.quantization.quant_utils import attribute_to_kwarg, ms_domain, QuantType
from onnx import onnx_pb as onnx_proto
'''
    Quantize LSTM
'''


class LSTMQuant(QuantOperatorBase): # pragma: no cover
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        '''
            parameter node: LSTM node.
            parameter new_nodes_list: List of new nodes created before processing this node.
            return: a list of nodes in topological order that represents quantized Attention node.
        '''
        node = self.node
        assert (node.op_type == "LSTM")

        if (not self.quantizer.is_valid_quantize_weight(node.input[1]) or
            not self.quantizer.is_valid_quantize_weight(node.input[2])):
            super().convert()
            return

        model = self.quantizer.model
        W = model.get_initializer(node.input[1])
        R = model.get_initializer(node.input[2])

        if (len(W.dims) != 3 or len(R.dims) != 3):
            super().convert()
            return

        [W_num_dir, W_4_hidden_size, W_input_size] = W.dims
        [R_num_dir, R_4_hidden_size, R_hidden_size] = R.dims

        if self.per_channel:
            del W.dims[0]
            del R.dims[0]
            W.dims[0] = W_num_dir * W_4_hidden_size
            R.dims[0] = R_num_dir * R_4_hidden_size

        quant_input_weight_tuple = self.quantizer.quantize_weight_per_channel(node.input[1], 
                                                        self.weight_dtype, self.weight_scheme, 0)
        quant_recurrent_weight_tuple = self.quantizer.quantize_weight_per_channel(node.input[2], 
                                                        self.weight_dtype, self.weight_scheme, 0)

        W_quant_weight = model.get_initializer(quant_input_weight_tuple[0])
        R_quant_weight = model.get_initializer(quant_recurrent_weight_tuple[0])

        W_quant_array = onnx.numpy_helper.to_array(W_quant_weight)
        R_quant_array = onnx.numpy_helper.to_array(R_quant_weight)

        W_quant_array = numpy.reshape(W_quant_array, (W_num_dir, W_4_hidden_size, W_input_size))
        R_quant_array = numpy.reshape(R_quant_array, (R_num_dir, R_4_hidden_size, R_hidden_size))

        W_quant_array = numpy.transpose(W_quant_array, (0, 2, 1))
        R_quant_array = numpy.transpose(R_quant_array, (0, 2, 1))

        W_quant_tranposed = onnx.numpy_helper.from_array(W_quant_array, \
                                                         quant_input_weight_tuple[0])
        R_quant_tranposed = onnx.numpy_helper.from_array(R_quant_array, 
                                                         quant_recurrent_weight_tuple[0])

        model.remove_initializers([W_quant_weight, R_quant_weight])
        model.add_initializer(W_quant_tranposed)
        model.add_initializer(R_quant_tranposed)

        W_quant_zp = model.get_initializer(quant_input_weight_tuple[1])
        R_quant_zp = model.get_initializer(quant_recurrent_weight_tuple[1])
        W_quant_scale = model.get_initializer(quant_input_weight_tuple[2])
        R_quant_scale = model.get_initializer(quant_recurrent_weight_tuple[2])

        if self.per_channel:
            W_quant_zp.dims[:] = [W_num_dir, W_4_hidden_size]
            R_quant_zp.dims[:] = [R_num_dir, R_4_hidden_size]
            W_quant_scale.dims[:] = [W_num_dir, W_4_hidden_size]
            R_quant_scale.dims[:] = [R_num_dir, R_4_hidden_size]

        inputs = []
        input_len = len(node.input)
        inputs.extend([node.input[0]])
        inputs.extend([quant_input_weight_tuple[0], quant_recurrent_weight_tuple[0]])
        inputs.extend([node.input[3] if input_len > 3 else ""])
        inputs.extend([node.input[4] if input_len > 4 else ""])
        inputs.extend([node.input[5] if input_len > 5 else ""])
        inputs.extend([node.input[6] if input_len > 6 else ""])
        inputs.extend([node.input[7] if input_len > 7 else ""])
        inputs.extend([quant_input_weight_tuple[2], 
                       quant_input_weight_tuple[1], 
                       quant_recurrent_weight_tuple[2], 
                       quant_recurrent_weight_tuple[1]])
                       
        kwargs = {}
        for attribute in node.attribute:
            kwargs.update(attribute_to_kwarg(attribute))
        kwargs["domain"] = ms_domain

        quant_lstm_name = node.name + "_quant"
        quant_lstm_node = onnx.helper.make_node("DynamicQuantizeLSTM", 
                                                inputs, node.output, quant_lstm_name, **kwargs)
        self.quantizer.remove_nodes.append(node)
        self.quantizer.new_nodes.append(quant_lstm_node)

class QDQLSTM(QDQOperatorBase): # pragma: no cover
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        return
 



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from .qdq_base_operator import QDQOperatorBase
from onnxruntime.quantization.quant_utils import QuantizedValueType, \
                                                 attribute_to_kwarg, ms_domain
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue


class QLinearActivation(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        if node.op_type in ['Relu', 'Clip']:
            return
            
        if len(self.quantizer.model.get_children(node)) == 0:
            return
        # No assert on op_type as it is controlled by registry
        # only try to quantize when given quantization parameters for it
        parent = self.quantizer.model.get_parents(node)[0]
        child = self.quantizer.model.get_children(node)[0]

        inputs = []
        inputs.extend(parent.input)
        inputs.extend(child.input[1:])

        qlinear_activation_output = child.output[0]
        kwargs = {}
        for attribute in node.attribute:
            kwargs.update(attribute_to_kwarg(attribute))
        kwargs["domain"] = ms_domain

        qlinear_activation_node = onnx.helper.make_node(
            "QLinear" + node.op_type, inputs,
            [qlinear_activation_output], node.name, **kwargs)

        self.quantizer.new_nodes.append(qlinear_activation_node)
        self.quantizer.remove_nodes.extend([parent, child, node])

class QDQRemovableActivation(QDQOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        if node.input[0] not in self.quantizer.quantized_value_map:
            return
        elif node.output[0] in [i.name for i in self.quantizer.model.model.graph.output]:
            self.quantizer.dequantize_tensor(node, node.input[0])
        else:
            self.quantizer.model.replace_input_of_all_nodes(node.output[0], node.input[0])
            self.quantizer.remove_nodes.append(node)

class QDQActivation(QDQOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        data_found, _, _, _, _ = self.quantizer._get_quantization_params(node.output[0])
        if not data_found:
            return
        super().quantize()
        node.name = node.name + "_quant"
